import torchio as tio
import nibabel as nib
import numpy as np
import matplotlib.pyplot as plt
import os
from pydactim.utils import get_name_of_path

def plot_histo(input_path):
    # Load your MRI image data into a numpy array
    mri_image = nib.load(input_path).get_fdata()

    # Flatten the 3D array to a 1D array
    image_flat = mri_image.flatten()

    # Plot the histogram
    plt.hist(image_flat, bins=256, alpha=0.75)
    plt.xlabel('Intensity')
    plt.ylabel('Frequency')
    plt.title('Histogram of MRI Image')
    plt.show()

def a_plot_histo(intput_data):
    # Flatten the 3D array to a 1D array
    image_flat = intput_data.flatten()

    # Plot the histogram
    plt.hist(image_flat, bins=256, range=(1, np.amax(intput_data)), alpha=0.75)
    plt.xlabel('Intensity')
    plt.ylabel('Frequency')
    plt.title('Histogram of MRI Image')
    plt.show()

def old_plot(*images):
    subject = tio.Subject(temp=tio.ScalarImage(images[0]))
    if len(images) > 1:
        for img in images:
            title = get_name_of_path(os.path.basename(img).replace(".nii.gz", ""))
            subject.add_image(tio.ScalarImage(img), title)
    else:
        title = get_name_of_path(os.path.basename(images[0]).replace(".nii.gz", ""))
        subject.add_image(tio.ScalarImage(images[0]), title)

    subject.remove_image("temp")
    subject.plot()

def plot(input_data, pixdim=None, slice=None, save=None):
    """ Plot image from a nifti file path or from an array.
        If using this function with an array, make sure to specify the pixdim value.

        Parameters
        ----------
        input_data : str/array
            Nifti file path or array that will be plotted

        pixdim : list, optional
            List of size 3 of the dimension of the voxels

        slice : int, optional
            The axial slice that will be displayed 
            
        save : str, optional
            The path to save the plot if defined

        """
    print(f"INFO - Plotting image")
    if type(input_data) is str:
        img = nib.load(input_data)
        data = img.get_fdata()
        pixdim = [round(dim, 2) for dim in img.header["pixdim"][1:4].astype(float)]
        title = f"{os.path.basename(input_data)} (shape={data.shape}, pixdim={pixdim})"
    else:
        data = input_data
        if pixdim is None:
            pixdim = [1,1,1] # Default pixdim
            print("WARNING - 'Pixdim' parameter is not defined, hence the pixel dimensions will be set to default (can deform the image)")
        if not isinstance(pixdim, list) or len(pixdim) != 3: raise ValueError(f"ERROR - The pixdim must be an instance of list with 3 elements, verify this:\n\t{pixdim :}")
        title = f"shape={data.shape}, pixdim={pixdim}"

    plt.style.use('dark_background')
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(14, 6))

    indices = np.array(data.shape) // 2
    
    i, j, k = indices
    if slice is not None: k = slice
    sag = np.rot90(np.fliplr(data[i, :, :]), -1)
    cor = np.rot90(np.fliplr(data[:, j, :]), -1)
    tra = np.rot90(np.fliplr(data[:, :, k]), -1)

    sag_aspect = pixdim[2] / pixdim[1]
    ax1.imshow(sag, aspect=sag_aspect, cmap="gray")
    ax1.set_title(f'Sag (slice={str(i)})', y=0.9)
    ax1.axis('off')

    cor_aspect = pixdim[2] / pixdim[0]
    ax2.imshow(cor, aspect=cor_aspect, cmap="gray")
    ax2.set_title(f'Cor (slice={str(j)})', y=0.9)
    ax2.axis('off')

    tra_aspect = pixdim[1] / pixdim[0]
    ax3.imshow(tra, aspect=tra_aspect, cmap="gray")
    ax3.set_title(f'Tra (slice={str(k)})', y=0.9)
    ax3.axis('off')

    fig.text(0.5, 0.95, title, horizontalalignment="center")
    if save is not None: plt.savefig(save)
    plt.tight_layout()
    plt.show()

class plot3D():
    def __init__(self, input_path):
        self.volume = nib.load(input_path)
        self.volume_data = self.check_orientation(self.volume)[::-1,:,::-1]
        self.multi_slice_viewer()

    def check_orientation(self, volume):
        x, y, z = nib.aff2axcodes(volume.affine)
        volume_data = volume.get_fdata()
        if x != 'R':
            volume_data = nib.orientations.flip_axis(volume_data, axis=0)
        if y != 'P':
            volume_data = nib.orientations.flip_axis(volume_data, axis=1)
        if z != 'S':
            volume_data = nib.orientations.flip_axis(volume_data, axis=2)
        return volume_data
        
    def multi_slice_viewer(self):
        plt.style.use('dark_background')
        fig, (ax1, ax2, ax3) = plt.subplots(1,3, figsize=[18,10])

        ax1.volume = np.transpose(self.volume_data, (0,2,1))
        ax1.index = ax1.volume.shape[0] // 2
        ax1.imshow(ax1.volume[ax1.index], cmap="gray", animated=True)

        ax2.volume = np.transpose(ax1.volume, (2,1,0))
        ax2.index = ax2.volume.shape[0] // 2
        ax2.imshow(ax2.volume[ax2.index], cmap="gray", animated=True)

        ax3.volume = np.transpose(ax1.volume, (1,2,0))
        ax3.index = ax3.volume.shape[0] // 2
        ax3.imshow(ax3.volume[ax3.index], cmap="gray", animated=True)

        fig.canvas.mpl_connect('scroll_event', self.process_scroll)
        fig.canvas.mpl_connect('button_press_event', self.process_key)

        ax1.set_axis_off()
        ax2.set_axis_off()
        ax3.set_axis_off()

        plt.tight_layout()
        plt.show()

    def process_scroll(self, event):
        fig = event.canvas.figure
        ax = event.inaxes
        if ax == None: return

        if event.button == 'down':
            self.next_slice(ax)
            print("down")
        elif event.button == 'up':
            self.previous_slice(ax)
            print("up")
        fig.canvas.draw()

    def process_key(self, event):
        fig = event.canvas.figure
        ax = event.inaxes
        if ax == None: return
        
        x = int(event.xdata)
        y = int(event.ydata)

        axs = plt.gcf().get_axes()
        axs.remove(ax)
        self.clicked_slice(axs, x, y)

        fig.canvas.draw()

    def previous_slice(self, ax):
        volume = ax.volume
        ax.index = (ax.index - 1) % volume.shape[0]
        ax.images[0].set_array((volume[ax.index]))

    def next_slice(self, ax):
        volume = ax.volume
        ax.index = (ax.index + 1) % volume.shape[0]
        ax.images[0].set_array((volume[ax.index]))

    def clicked_slice(self, axs, x, y):
        axs[0].images[0].set_array((axs[0].volume[x]))
        axs[1].images[0].set_array((axs[1].volume[y]))



import os
import SimpleITK as sitk
import torchio as tio
import nibabel as nib
import numpy as np
from skimage import morphology
import matplotlib.pyplot as plt
from dipy.segment.tissue import TissueClassifierHMRF
import itk
from .brain_extraction import run_hd_bet

class Dactim():
    def __init__(self):
        pass

    def init(self, status):
        self.status = status
        self.precise_path = ""
        self.exit = False

        self.nii_path = None
        self.nii = None
        self.nii_data = None
        self.nii_mask = None
        self.nii_mask_data = None
        self.nii_ref = None
        self.nii_ref_data = None

        self.output_path = None
        self.mask_path = None

    def load_data(self, input_data, atlas_data=False, mask_data=False):
        if type(input_data) is str:
            self.nii_path = input_data
            self.precise_path = f"for\n\t{input_data :}"
            self.nii = nib.load(input_data)
            self.nii_data = self.nii.get_fdata()
        else:
            self.nii_data = input_data

        if type(atlas_data) is str:
            self.nii_atlas_path = atlas_data
            self.precise_path = f"\n\t{atlas_data :}"
            self.nii_atlas = nib.load(atlas_data)
            self.nii_atlas_data = self.nii_atlas.get_fdata()
        else:
            self.nii_atlas_data = atlas_data

        if type(mask_data) is str:
            self.nii_mask_path = mask_data
            self.precise_path = f"\n\t{mask_data :}"
            self.nii_mask = nib.load(mask_data)
            self.nii_mask_data = self.nii_mask.get_fdata()
        else:
            self.nii_mask_data = mask_data

    def generate_output_path(self, input_path, suffix, force, other=None, other_format=None):
        self.output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
        if other is not None:
            self.other_path = self.output_path.replace(".nii.gz", "_"+other+"."+other_format)
        if os.path.exists(self.output_path) and not force:
            print(f"INFO - {self.status} already done for\n\t{input_path :}")
            if other is not None:
                if os.path.exists(self.other_path):
                    self.exit = True

    def nib_save(self, extra=False, extra_data=None):
        print(f"INFO - Saving generated image at\n\t{self.output_path :}")
        nib.save(nib.Nifti1Image(self.output_data, self.nii.affine), self.output_path)
        if extra is False:
            return self.output_path
        else:
            print(f"INFO - Saving generated image at\n\t{self.output_path :}")
            nib.save(nib.Nifti1Image(extra_data, self.nii.affine), self.other_path)
            return self.output_path, self.other_path

    def sitk_save(self, extra=False, extra_data=None):
        print(f"INFO - Saving generated image at\n\t{self.output_path :}")
        sitk.WriteImage(self.output_data, self.output_path)
        if extra is False:
            return self.output_path
        else:
            print(f"INFO - Saving generated image at\n\t{self.output_path :}")
            sitk.WriteImage(extra_data, self.other_path)
            return self.output_path, self.other_path

    def itk_save(self, extra=False, extra_data=None):
        print(f"INFO - Saving generated image at\n\t{self.output_path :}")
        itk.imwrite(self.output_data, self.output_path)
        if extra is False:
            return self.output_path
        else:
            print(f"INFO - Saving generated image at\n\t{self.output_path :}")
            itk.imwrite(extra_data, self.other_path)
            return self.output_path, self.other_path

    def nib_to_sitk(self, data):
        # sitk is (x y z) while np is (z y x)
        data = np.transpose(data, (2,1,0)) 
        sitk_data = sitk.GetImageFromArray(data)
        # sitk_data.SetOrigin((0, 0, 0))
        return sitk_data

    def plot(self, input_data, pixdim=None, slice=None):
        print(f"INFO - Plotting image")
        if type(input_data) is str:
            img = nib.load(input_data)
            data = img.get_fdata()
            pixdim = img.header["pixdim"][1:4].astype(float)
            title = f"{os.path.basename(input_data)} (shape={data.shape}, pixdim={pixdim})"
        else:
            data = input_data
            if pixdim is None:
                pixdim = [1,1,1] # Default pixdim
                print("WARNING - 'Pixdim' parameter is not defined, hence the pixel dimensions will be set to default (can deform the image)")
            title = f"shape={data.shape}, pixdim={pixdim}"

        plt.style.use('dark_background')
        fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(14, 6))

        indices = np.array(data.shape) // 2
       
        i, j, k = indices
        if slice is not None: k = slice
        sag = np.rot90(np.fliplr(data[i, :, :]), -1)
        cor = np.rot90(np.fliplr(data[:, j, :]), -1)
        tra = np.rot90(np.fliplr(data[:, :, k]), -1)

        sag_aspect = pixdim[2] / pixdim[1]
        ax1.imshow(sag, aspect=sag_aspect, cmap="gray")
        ax1.set_title(f'Sag (slice={str(i)})', y=0.9)
        ax1.axis('off')

        cor_aspect = pixdim[2] / pixdim[0]
        ax2.imshow(cor, aspect=cor_aspect, cmap="gray")
        ax2.set_title(f'Cor (slice={str(j)})', y=0.9)
        ax2.axis('off')

        tra_aspect = pixdim[1] / pixdim[0]
        ax3.imshow(tra, aspect=tra_aspect, cmap="gray")
        ax3.set_title(f'Tra (slice={str(k)})', y=0.9)
        ax3.axis('off')

        fig.text(0.5, 0.95, title, horizontalalignment="center")
        # plt.savefig(r'E:\Leo\data\sub-001\ses-01\anat\myfig.png')
        plt.tight_layout()
        plt.show()

    def skull_stripping(self, input_data, model_path, mask=True, force=True, suffix="brain"):
        """ HD-BET to extract the brain of any anatomic mri image.
        This script is a slightly modified version of the HD-BET script that can be found at the following link:
        https://github.com/MIC-DKFZ/HD-BET

        Isensee F, Schell M, Tursunova I, Brugnara G, Bonekamp D, Neuberger U, Wick A, Schlemmer HP, Heiland S, Wick W,
        Bendszus M, Maier-Hein KH, Kickingereder P. Automated brain extraction of multi-sequence MRI using artificia
        neural networks. arXiv preprint arXiv:1901.11341, 2019.
        https://doi.org/10.1002/hbm.24750

        Parameters
        ----------
        input_data : str/array
            The image in which the brain will be extracted. Can be a path or an array

        model_path : str
            Path of the directory containing the folds of the model. 
            The files can be downloaded from the following path: 
            https://zenodo.org/record/2540695

        mask : bool
            To get the binary mask of the brain
            
        suffix : str
            Nifti file suffixe for the new generated image

        """

        if os.path.exists(model_path) and len(os.listdir(model_path)) == 0:
            raise FileNotFoundError("ERROR - No fold found in the specific directory. You must download the folds of the model at the following link: https://zenodo.org/record/2540695")
        elif not os.path.exists(model_path):
            raise NotADirectoryError("ERROR - No valid directory for the model folds")

        self.init("Brain extraction")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force, other="brain_mask", other_format="nii.gz") 
            if self.exit and mask: return self.output_path, self.other_path
            elif self.exit: return self.output_path

        self.load_data(input_data)
        print(f"INFO - Starting the brain extraction {self.precise_path}")

        self.nii_data = self.nib_to_sitk(self.nii_data)

        self.output_data, self.nii_mask_data = run_hd_bet(self.nii_data, model_path=model_path)
        self.output_data = np.transpose(self.output_data, (2,1,0)) 
        self.nii_mask_data = np.transpose(self.nii_mask_data, (2,1,0)) 

        if type(input_data) is str: return self.nib_save(extra=mask, extra_data=self.nii_mask_data)
        else: 
            if mask: return self.output_data, self.nii_mask_data
            else: return self.output_data

    def n4_bias_field_correction(self, input_data, pixdim=None, mask=False, force=True, suffix="corrected"):
        """ Correct the bias field.

        Parameters
        ----------
        input_data : str/array
            Nifti file path or array that will be corrected

        pixdim : array, optional
            Array of size 3 of the dimension of the voxels

        mask : bool, optional
            Get the mask of the bias field
            
        suffix : str
            Nifti file suffixe for the new generated image

        """
        self.init("Bias field correction")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force, other="bias_field", other_format="nii.gz") 
            if self.exit and mask: return self.output_path, self.other_path
            elif self.exit: return self.output_path

        self.load_data(input_data)
        print(f"INFO - Starting bias field correction {self.precise_path}")

        self.nii_data = self.nib_to_sitk(self.nii_data)
        if pixdim is not None: self.nii_data.SetSpacing(pixdim.astype(float))

        # nifti = sitk.ReadImage(input_path, sitk.sitkFloat32)
        mask_otsu = sitk.OtsuThreshold(self.nii_data,0,1,200)
        corrector = sitk.N4BiasFieldCorrectionImageFilter()

        self.output_data = sitk.GetArrayFromImage(corrector.Execute(self.nii_data, mask_otsu))
        log_bias_field = sitk.GetArrayFromImage(corrector.GetLogBiasFieldAsImage(self.nii_data))
        self.output_data = np.transpose(self.output_data, (2,1,0))
        log_bias_field = np.transpose(log_bias_field, (2,1,0))

        if type(input_data) is str: return self.nib_save(extra=mask, extra_data=log_bias_field)
        else: 
            if mask: return self.output_data, log_bias_field
            else: return self.output_data

    def tissue_classifier(self, input_data, pve=True, force=True, suffix="fast"):
        """ Normalize image between 0 and 1.

        Parameters
        ----------
        input_data : str/array
            The image that will be normalized. Can be a path or an array

        suffix : str
            Nifti file suffixe for the new generated image
        """
        self.init("Tissue classification")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force, other="pve", other_format="nii.gz") 
            if self.exit and pve: return self.output_path, self.other_path
            elif self.exit: return self.output_path

        self.load_data(input_data)
        print(f"INFO - Starting tissue classification {self.precise_path}")
        
        nclass = 3
        beta = 0.1

        hmrf = TissueClassifierHMRF(verbose=False)
        _, self.output_data, self.nii_mask_data = hmrf.classify(self.nii_data, nclass, beta)

        if type(input_data) is str: return self.nib_save(extra=pve, extra_data=self.nii_mask_data)
        else: 
            if pve: return self.output_data, self.nii_mask_data
            else: return self.output_data

    def registration(self, atlas_data, input_data, matrix=False, elastix=True, force=True, suffix="flirt"):
        """ Register a nifti image to an atlas.

        Parameters
        ----------
        atlas_data : str
            Atlas that will be used as fixed image

        input_data : str
            Image that will be registered

        matrix : bool, optional
            Get the transformation matrix

        elastix : bool, optional
            Make the registration elastix

        suffix : str
            Nifti file suffixe for the new generated image

        """
        self.init("Registeration")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force, other="matrix", other_format="mat") 
            if self.exit and matrix: return self.output_path, self.other_path
            elif self.exit: return self.output_path

        self.load_data(input_data, atlas_data=atlas_data)
        print(f"INFO - Starting registration {self.precise_path}")

        if elastix:
            fixed = itk.imread(atlas_data, itk.F)
            moving = itk.imread(input_data, itk.F)

            self.output_data, transformation_matrix = itk.elastix_registration_method(fixed, moving)
            
            if type(input_data) is str: return self.itk_save(extra=matrix, extra_data=transformation_matrix)
            else: 
                if matrix: return sitk.GetArrayFromImage(self.output_data), transformation_matrix
                else: return sitk.GetArrayFromImage(self.output_data)
        else:
            fixed = sitk.ReadImage(atlas_data, sitk.sitkFloat32)
            moving = sitk.ReadImage(input_data, sitk.sitkFloat32)

            initial_transform = sitk.CenteredTransformInitializer(
                fixed, 
                moving, 
                sitk.Euler3DTransform(), 
                sitk.CenteredTransformInitializerFilter.GEOMETRY
            )

            R = sitk.ImageRegistrationMethod()
            R.SetMetricAsMeanSquares()
            R.SetInterpolator(sitk.sitkLinear)
            R.SetOptimizerAsGradientDescent(learningRate=1.0, numberOfIterations=1000, convergenceMinimumValue=1e-8, convergenceWindowSize=10)
            R.SetOptimizerScalesFromPhysicalShift()
            R.SetShrinkFactorsPerLevel(shrinkFactors = [4,2,1])
            R.SetSmoothingSigmasPerLevel(smoothingSigmas=[2,1,0])
            R.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

            R.SetInitialTransform(initial_transform, inPlace=False)
            transformation_matrix = R.Execute(fixed, moving)

            self.output_data = sitk.Resample(moving, fixed, transformation_matrix, sitk.sitkLinear, 0.0, moving.GetPixelID())
            self.output_data = sitk.GetArrayFromImage(self.output_data)

            if type(input_data) is str: return self.sitk_save(extra=matrix, extra_data=transformation_matrix)
            else: 
                if matrix: return self.output_data, transformation_matrix
                else: return self.output_data

    def histogram_matching(self, atlas_data, input_data, force=True, suffix="hm"):
        """ Match the histogram of two images.

        Parameters
        ----------
        atlas_data : str/array
            The fixed image on which the other image will be matched. Can be a path or an array

        input_data : str/array
            The moving image that needs to be filtered. Can be a path or an array

        suffix : str
            Nifti file suffixe for the new generated image
        """
        self.init("Histogram matching")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force) 
            if self.exit: return self.output_path

        self.load_data(input_data, atlas_data=atlas_data)
        print(f"INFO - Starting histogram matching for {self.precise_path}")

        him = sitk.HistogramMatchingImageFilter()
        him.SetThresholdAtMeanIntensity(True)

        self.nii_data = self.nib_to_sitk(self.nii_data)
        self.nii_atlas_data = self.nib_to_sitk(self.nii_atlas_data)

        self.output_data = sitk.GetArrayFromImage(him.Execute(self.nii_data, self.nii_atlas_data))
        self.output_data = np.transpose(self.output_data, (2,1,0))

        if type(input_data) is str: return self.nib_save()
        else: return self.output_data

    def normalize(self, input_data, force=True, suffix="minmax"):
        """ Normalize image between 0 and 1.

        Parameters
        ----------
        input_data : str/array
            The image that will be normalized. Can be a path or an array

        suffix : str
            Nifti file suffixe for the new generated image
        """
        self.init("Normalization")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force) 
            if self.exit: return self.output_path

        self.load_data(input_data)
        print(f"INFO - Starting normalization {self.precise_path}")

        self.output_data = (self.nii_data - self.nii_data.min()) / (self.nii_data.max() - self.nii_data.min())
        
        if type(input_data) is str: return self.nib_save()
        else: return self.output_data
        
    def variance(self, input_data, offset=9, force=True, suffix="var"):
        """ Amplify borders by creating a variance map.

        Parameters
        ----------
        input_data : str/array
            The image that will be used to create the variance map. Can be a path or an array

        offset : int
            offset of the variance, the highest it is, the strongest the borders

        suffix : str
            Nifti file suffixe for the new generated image
        """
        self.init("Variance map")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force) 
            if self.exit: return self.output_path

        self.load_data(input_data)
        print(f"INFO - Starting variance map {self.precise_path}")

        var_data = np.zeros_like(self.nii_data)
        for x in range(self.nii_data.shape[0]):
            for y in range(self.nii_data.shape[1]):
                for z in range(self.nii_data.shape[2]):
                    if self.nii_data[x,y,z] != 0:
                        variance = np.var(self.nii_data[x-offset//2+1:x+offset//2+1,y-offset//2+1:y+offset//2+1,z-offset//2+1:z+offset//2+1])
                        var_data[x,y,z] = variance

        if type(input_data) is str: return self.nib_save()
        else: return self.output_data
    
    def crop_mri(self, input_data, force=True, suffix="cropped"):
        """ Crop the maximum of noise in an image.

        Parameters
        ----------
        input_data : str/array
            The image that will bet cropped. Can be a path or an array

        suffix : str
            Nifti file suffixe for the new generated image

        Returns
        -------
        cropped_idx : list if int
            List of the indices that was used to crop the input image

        """
        self.init("Automatic crop")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force) 
            if self.exit: return self.output_path

        self.load_data(input_data)
        print(f"INFO - Starting automatic crop {self.precise_path}")
        
        z_dim, y_dim, x_dim = self.nii_data.shape

        # Calculate the threshold for the image intensity to determine the background
        threshold = np.mean(self.nii_data) + 2 * np.std(self.nii_data)

        # Find the maximum intensity value within a margin around the center of the image
        z_margin, y_margin, x_margin = z_dim // 4, y_dim // 4, x_dim // 4
        z_center, y_center, x_center = z_dim // 2, y_dim // 2, x_dim // 2
        center_intensity = np.max(self.nii_data[z_center - z_margin:z_center + z_margin,
                                            y_center - y_margin:y_center + y_margin,
                                            x_center - x_margin:x_center + x_margin])

        # Set a lower threshold to ensure that the cube contains some useful information
        lower_threshold = center_intensity * 0.75

        # Determine the starting and ending indices for each axis
        z_start = np.argmax(np.max(self.nii_data, axis=(1, 2)) > threshold)
        z_end = z_dim - np.argmax(np.max(self.nii_data[::-1, :, :], axis=(1, 2)) > threshold)
        y_start = np.argmax(np.max(self.nii_data, axis=(0, 2)) > threshold)
        y_end = y_dim - np.argmax(np.max(self.nii_data[:, ::-1, :], axis=(0, 2)) > threshold)
        x_start = np.argmax(np.max(self.nii_data, axis=(0, 1)) > threshold)
        x_end = x_dim - np.argmax(np.max(self.nii_data[:, :, ::-1], axis=(0, 1)) > threshold)

        # Make sure the cube contains some useful information
        while (self.nii_data[z_start:z_end, y_start:y_end, x_start:x_end].max() < lower_threshold):
            z_start += 1
            z_end -= 1
            y_start += 1
            y_end -= 1
            x_start += 1
            x_end -= 1

        # Crop the MRI image to extract the cube
        self.output_data = self.nii_data[z_start:z_end, y_start:y_end, x_start:x_end]
        cropped_idx = [z_start, z_end, y_start, y_end, x_start, x_end]

        if type(input_data) is str: return self.nib_save(), cropped_idx
        else: return self.output_data, cropped_idx

    def apply_crop(self, input_data, crop, force=True, suffix="cropped"):
        """ Crop the image according crop indices.

        Parameters
        ----------
        input_data : str/array
            The image that will be cropped. Can be a path or an array

        crop : list
            List of 6 integers to determine the cropped indices 

        suffix : str
            Nifti file suffixe for the new generated image
        """
        self.init("Applying crop")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force) 
            if self.exit: return self.output_path

        self.load_data(input_data)
        print(f"INFO - Starting crop with shape {crop} {self.precise_path}")

        self.output_data = self.nii_data[crop[0]:crop[1], crop[2]:crop[3], crop[4]:crop[5]]

        if type(input_data) is str: return self.nib_save()
        else: return self.output_data

    def remove_small_object(self, input_data, min_size, force=False, suffix="opened"):
        """ Remove small objects above a minimum size

        Parameters
        ----------
        input_data : str/array
            Nifti file path of the binary mask in which will be removed all the small objects. Can be a path or an array

        min_size : int
            Minimum size of objects to be removed

        suffix : str
            Nifti file suffixe for the new generated image

        """
        self.init("Remove small objects")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force) 
            if self.exit: return self.output_path

        self.load_data(input_data)
        print(f"INFO - Starting to remove small objects {self.precise_path}")
        
        self.output_data = morphology.remove_small_objects(self.nii_data.astype(bool), min_size=min_size).astype(np.int16)
    
        if type(input_data) is str: return self.nib_save()
        else: return self.output_data

    def substract(self, atlas_data, input_data, force=True, suffix="sub"):
        """ Match the histogram of two images.

        Parameters
        ----------
        atlas_data : str/array
            The image on which the other image will be substracted. Can be a path or an array

        input_data : str/array
            The fixed image that will be substracted. Can be a path or an array

        suffix : str
            Nifti file suffixe for the new generated image
        """
        self.init("Substract")
        if type(input_data) is str:
            self.generate_output_path(input_data, suffix, force) 
            if self.exit: return self.output_path

        self.load_data(input_data, atlas_data=atlas_data)
        print(f"INFO - Starting to substract between {self.precise_path}")

        self.output_data = self.nii_atlas_data - self.nii_data

        if type(input_data) is str: return self.nib_save()
        else: return self.output_data


from PySide6.QtWidgets import QMessageBox, QGridLayout, QWidget, QLabel
from PySide6.QtGui import QImage, QPixmap, QAction
from PySide6.QtCore import Qt, QEvent
import pyqtgraph as pg

import nibabel as nib
import numpy as np

from .custom import CustomLabel
from .utils import is_dsc_param, create_lut, reset_layout, to_sagittal, to_coronal, to_axial, create_4D_grid, to_3D
from .settings import *
from scipy import ndimage

class ViewData():
    def __init__(self, data, slice, pixdim, z):
        self.data = data
        self.slice = slice
        self.pixdim = pixdim
        self.z = z

class ViewPanel(QWidget):
    def __init__(self, path, update_text, plot_aif, update_roi, state="3D"):
        super().__init__()

        self.layout = QGridLayout()
        self.setLayout(self.layout)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setStyleSheet("background-color: #000000")

        self.state = state

        if is_dsc_param(path):
            self.lut = "Rainbow"
        else:
            self.lut = "Grayscale"
        
        self.update_text = update_text
        self.plot_aif = plot_aif
        self.update_roi = update_roi
        self.init(path)

    def init(self, path):
        reset_layout(self.layout)

        self.path = path
        self.load_data(self.path)

        if is_dsc_param(self.path):
            self.lut = "Rainbow"
        else:
            self.lut = "Grayscale"

        if not self.ensure3D:
            self.sagittal_label = CustomLabel(self.path, self.s, self.update_slice, self.update_label_aif, self.update_label_roi, "Sagittal", size=(500,500))
            self.sagittal_label.lut = create_lut(self.lut)
            self.coronal_label = CustomLabel(self.path, self.c, self.update_slice, self.update_label_aif, self.update_label_roi, "Coronal", size=(500,500))
            self.coronal_label.lut = create_lut(self.lut)
            self.axial_label = CustomLabel(self.path, self.a, self.update_slice, self.update_label_aif, self.update_label_roi, "Axial", size=(500,500))
            self.axial_label.lut = create_lut(self.lut)

        if self.ensure3D:
            self.axial_label = CustomLabel(self.path, self.a, self.update_slice, self.update_label_aif, self.update_label_roi, "Axial", size=(1000,1000))
            self.axial_label.lut = create_lut(self.lut)
        
        self.update_layout()

    def load_data(self, path):
        self.img = nib.load(path)
        self.data = np.array(self.img.dataobj, dtype=np.float32)
        self.shape = self.img.shape
        self.pixdim = self.img.header["pixdim"][1:4]

        self.mosaic = False
        self.mosaic_data = None
        self.max = np.max(self.data)
        # self.plot_aif.setRange(yRange=[0, self.max])
        # self.data = ndimage.uniform_filter(self.data, size=3)
        # self.data = ndimage.gaussian_filter(self.data, sigma=3)


        self.ensure3D = False
        if len(self.shape) == 4:
            self.ensure3D = True
            self.state = "Axial"
            self.offset = self.shape[2]
            self.a_slice = self.shape[2] // 2
            self.adata, self.apixdim, self.az = to_3D(self.data, self.pixdim)  
            self.a = ViewData(self.adata[..., self.a_slice], self.a_slice, self.apixdim, self.az)
        else:
            self.offset = 1

            self.sdata, self.s_slice, self.spixdim, self.sz = to_sagittal(self.data, self.pixdim, path)
            self.s = ViewData(self.sdata[..., self.s_slice], self.s_slice, self.spixdim, self.sz)

            self.cdata, self.c_slice, self.cpixdim, self.cz = to_coronal(self.data, self.pixdim, path)
            self.c = ViewData(self.cdata[..., self.c_slice], self.c_slice, self.cpixdim, self.cz)

            self.adata, self.a_slice, self.apixdim, self.az = to_axial(self.data, self.pixdim, path)
            self.a = ViewData(self.adata[..., self.a_slice], self.a_slice, self.apixdim, self.az)

    def get_state_label(self):
        if self.state == "Sagittal": return self.sagittal_label
        elif self.state == "Coronal": return self.coronal_label
        elif self.state == "Axial": return self.axial_label

    def update_labels(self):
        if not self.ensure3D:
            self.sagittal_label.data = self.sdata[..., self.s_slice]
            self.sagittal_label.pixdim = self.spixdim

            self.coronal_label.data = self.cdata[..., self.c_slice]
            self.coronal_label.pixdim = self.cpixdim

        self.axial_label.data = self.adata[..., self.a_slice]
        self.axial_label.pixdim = self.apixdim

    def update_slice(self, idx, axis):
        if axis == "Sagittal":
            self.s_slice = idx
            self.sagittal_label.slice_index = self.s_slice
            self.sagittal_label.data = self.sdata[..., self.s_slice]
            if self.sagittal_label.overlay: self.sagittal_label.overlay_data = self.osdata[..., self.s_slice]
        elif axis == "Coronal":
            self.c_slice = idx
            self.coronal_label.slice_index = self.c_slice
            self.coronal_label.data = self.cdata[..., self.c_slice]
            if self.coronal_label.overlay: self.coronal_label.overlay_data = self.ocdata[..., self.c_slice]
        elif axis == "Axial":
            self.a_slice = idx
            self.axial_label.slice_index = self.a_slice
            if self.mosaic is True: self.axial_label.data = self.mosaic_data[..., self.a_slice]
            else: self.axial_label.data = self.adata[..., self.a_slice]
            if self.axial_label.overlay: self.axial_label.overlay_data = self.oadata[..., self.a_slice]
        self.update_text()
    
    def update_layout(self):
        if self.state == "3D":
            self.layout.addWidget(self.sagittal_label, 0, 1)
            self.layout.addWidget(self.coronal_label, 0, 0)
            self.layout.addWidget(self.axial_label, 1, 0)
            for i in reversed(range( self.layout.count())):
                label = self.layout.itemAt(i).widget()
                label.setMinimumSize(THREED_WIDTH, THREED_HEIGHT)
                label.state = self.state
                label.update_image()
        else:
            if self.state == "Sagittal":
                self.layout.addWidget(self.sagittal_label, 0,0)
            elif self.state == "Coronal":
                self.layout.addWidget(self.coronal_label, 0,0)
            elif self.state == "Axial":
                if len(self.shape) == 4:
                    if self.mosaic is True:
                        if self.mosaic_data is None: self.mosaic_data = create_4D_grid(self.data)
                        self.axial_label.mosaic = True
                        self.axial_label.data = self.mosaic_data[...,0]
                        self.a_slice = 0
                        self.axial_label.slice_index = self.a_slice
                        self.axial_label.z = self.data.shape[3]
                        self.offset = 1
                    else:
                        self.axial_label.data = self.adata[...,0]
                        self.axial_label.mosaic = False
                        self.a_slice = self.adata.shape[2] // 2
                        self.axial_label.slice_index = self.a_slice
                        self.axial_label.z = self.adata.shape[2]
                        self.offset = self.data.shape[2]

                self.layout.addWidget(self.axial_label, 0,0)

            label = self.layout.itemAt(0).widget()
            label.setMinimumSize(ONED_WIDTH, ONED_HEIGHT)
            label.state = self.state
            label.update_image()

    def update_label_aif(self):
        self.update_text()
        if ("perf." in self.path or "t2star." in self.path or "t1w." in self.path) and len(self.shape) == 4:
            x = min((self.axial_label.roi_x + (self.axial_label.roi_width // 2)) * self.shape[1] / ONED_WIDTH, self.shape[1]-1)
            y = min((self.axial_label.roi_y + (self.axial_label.roi_width // 2)) * self.shape[0] / ONED_WIDTH, self.shape[0]-1)
            z =  self.a_slice % self.shape[2]
            temp = np.rot90(self.data)
            aif = temp[int(y), int(x), z, :]

            idx = np.where(np.max(self.data[:,:,z,:], axis=-1) > 1000)
            mean = np.mean(self.data[idx[0], idx[1], z, :], axis=0)

            self.plot_aif.clear()
            self.plot_aif.plot(aif, pen=pg.mkPen(color=(255, 0, 0)))
            self.plot_aif.plot(mean, pen=pg.mkPen(color=(255, 255, 255)))
            self.v_line = pg.InfiniteLine(angle=90, movable=False, pen=pg.mkPen(color=(100, 100, 255)))
            self.v_line.setPos(self.a_slice // self.shape[2])
            self.plot_aif.addItem(self.v_line)
        else:
            aif = None
            mean = None

    def update_label_roi(self):
        self.update_text()
        if (is_dsc_param(self.path) or is_dsc_param(self.axial_label.overlay_path)) and len(self.shape) == 3:
            self.plot_aif.clear()
            histo, bins = np.histogram(self.axial_label.roi, bins=49)
            # print(histo)
            self.plot_aif.plot(histo, pen=pg.mkPen(color=(255, 0, 0)))
            self.update_roi(
                self.axial_label.roi_xx, 
                self.axial_label.roi_yy, 
                self.axial_label.slice_index,
                self.axial_label.roi_mean,
                self.axial_label.roi_std,
                self.axial_label.roi_min,
                self.axial_label.roi_max,
                self.axial_label.roi_mm3,
            )
            # self.plot_aif.plot(mean, pen=pg.mkPen(color=(255, 255, 255)))
        
        # self.plot_aif(aif)

    def change_lut(self, lut):
        self.lut = lut
        color_lut = create_lut(lut)
        for i in reversed(range(self.layout.count())):
            label = self.layout.itemAt(i).widget()
            label.lut = color_lut
            label.update_image()

    def change_overlay_lut(self, lut):
        self.overlay_lut = lut
        color_lut = create_lut(lut)
        for i in reversed(range(self.layout.count())):
            label = self.layout.itemAt(i).widget()
            label.overlay_lut = color_lut
            label.update_image()

    def add_overlay(self, path):
        print("INFO - Loading overlay")
        self.opath = path
        self.oimg = nib.load(self.opath)
        self.odata = self.oimg.get_fdata()
        self.oshape = self.oimg.shape
        self.opixdim = self.oimg.header["pixdim"][1:4]

        self.osdata, _, self.ospixdim, _ = to_sagittal(self.odata, self.opixdim, self.opath)
        self.ocdata, _, self.ocpixdim, _ = to_coronal(self.odata, self.opixdim, self.opath)
        self.oadata, _, self.oapixdim, _ = to_axial(self.odata, self.opixdim, self.opath)

        # If not in the same space, alert user
        if self.oshape != self.shape:
            dlg = QMessageBox(self)
            dlg.setWindowTitle("Warning")
            dlg.setIcon(QMessageBox.Icon.Warning)
            dlg.setText("The overlay data is not in the\nsame space than the previous data")
            button = dlg.exec()

        if is_dsc_param(self.opath):
            lut = "Rainbow"
        else:
            lut = "Grayscale"
        self.sagittal_label.set_overlay(path, self.osdata[..., self.s_slice], self.ospixdim, lut)
        self.coronal_label.set_overlay(path, self.ocdata[..., self.c_slice], self.ocpixdim, lut)
        self.axial_label.set_overlay(path, self.oadata[..., self.a_slice], self.oapixdim, lut)
        print("INFO - Overlay successfully loaded")

    def keyPressEvent(self, event):
        if event.type() == QEvent.Type.KeyPress:
            if event.key() == Qt.Key_S and not self.ensure3D:
                self.state = "Sagittal"
                reset_layout(self.layout)
                self.update_layout()
                self.update_text()
            elif event.key() == Qt.Key_C and not self.ensure3D:
                self.state = "Coronal"
                reset_layout(self.layout)
                self.update_text()
                self.update_layout()
            elif event.key() == Qt.Key_A:
                self.state = "Axial"
                self.mosaic = False
                reset_layout(self.layout)
                self.update_layout()
                self.update_text()
            elif event.key() == Qt.Key_F and not self.ensure3D:
                self.state = "3D"
                reset_layout(self.layout)
                self.update_layout()
                self.update_text()
            elif event.key() == Qt.Key_M:
                self.mosaic = True
                reset_layout(self.layout)
                self.update_layout()
                self.update_text()
            elif self.ensure3D:
                if event.key() == Qt.Key_Right:
                    new_index = self.a_slice + self.offset
                    if self.mosaic:
                        if new_index > self.data.shape[3] - 1: new_index -= self.offset
                    else:
                        if new_index > self.adata.shape[2] - 1: new_index -= self.offset
                    self.update_slice(new_index, "Axial")
                    self.v_line.setPos(self.a_slice // self.shape[2])
                    self.axial_label.update_image()
                elif event.key() == Qt.Key_Left:
                    new_index = max(self.a_slice - self.offset, self.a_slice % self.offset)
                    self.update_slice(new_index, "Axial")
                    self.v_line.setPos(self.a_slice // self.shape[2])
                    self.axial_label.update_image()

        return super().keyPressEvent(event)


"""
https://github.com/MIC-DKFZ/HD-BET

Isensee F, Schell M, Tursunova I, Brugnara G, Bonekamp D, Neuberger U, Wick A, Schlemmer HP, Heiland S, Wick W, Bendszus M, Maier-Hein KH, Kickingereder P. 
Automated brain extraction of multi-sequence MRI using artificial neural networks. 
Hum Brain Mapp. 2019; 1–13. 
https://doi.org/10.1002/hbm.24750

"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import SimpleITK as sitk
from skimage.transform import resize
from skimage.morphology import label
from abc import abstractmethod
import os


""" UTILS """
def get_params_fname(model_path, fold):
    return os.path.join(model_path, "%d.model" % fold)

def softmax_helper(x):
    rpt = [1 for _ in range(len(x.size()))]
    rpt[1] = x.size(1)
    x_max = x.max(1, keepdim=True)[0].repeat(*rpt)
    e_x = torch.exp(x - x_max)
    return e_x / e_x.sum(1, keepdim=True).repeat(*rpt)

def postprocess_prediction(seg):
    # basically look for connected components and choose the largest one, delete everything else
    print("running postprocessing... ")
    mask = seg != 0
    lbls = label(mask, connectivity=mask.ndim)
    lbls_sizes = [np.sum(lbls == i) for i in np.unique(lbls)]
    largest_region = np.argmax(lbls_sizes[1:]) + 1
    seg[lbls != largest_region] = 0
    return seg

class SetNetworkToVal(object):
    def __init__(self, use_dropout_sampling=False, norm_use_average=True):
        self.norm_use_average = norm_use_average
        self.use_dropout_sampling = use_dropout_sampling

    def __call__(self, module):
        if isinstance(module, nn.Dropout3d) or isinstance(module, nn.Dropout2d) or isinstance(module, nn.Dropout):
            module.train(self.use_dropout_sampling)
        elif isinstance(module, nn.InstanceNorm3d) or isinstance(module, nn.InstanceNorm2d) or \
                isinstance(module, nn.InstanceNorm1d) \
                or isinstance(module, nn.BatchNorm2d) or isinstance(module, nn.BatchNorm3d) or \
                isinstance(module, nn.BatchNorm1d):
            module.train(not self.norm_use_average)

""" CONFIG """
class BaseConfig(object):
    def __init__(self):
        pass

    @abstractmethod
    def get_split(self, fold, random_state=12345):
        pass

    @abstractmethod
    def get_network(self, mode="train"):
        pass

    @abstractmethod
    def get_basic_generators(self, fold):
        pass

    @abstractmethod
    def get_data_generators(self, fold):
        pass

    def preprocess(self, data):
        return data

    def __repr__(self):
        res = ""
        for v in vars(self):
            if not v.startswith("__") and not v.startswith("_") and v != 'dataset':
                res += (v + ": " + str(self.__getattribute__(v)) + "\n")
        return res

class HD_BET_Config(BaseConfig):
    def __init__(self):
        super(HD_BET_Config, self).__init__()

        self.EXPERIMENT_NAME = self.__class__.__name__ # just a generic experiment name

        # network parameters
        self.net_base_num_layers = 21
        self.BATCH_SIZE = 2
        self.net_do_DS = True
        self.net_dropout_p = 0.0
        self.net_use_inst_norm = True
        self.net_conv_use_bias = True
        self.net_norm_use_affine = True
        self.net_leaky_relu_slope = 1e-1

        # hyperparameters
        self.INPUT_PATCH_SIZE = (128, 128, 128)
        self.num_classes = 2
        self.selected_data_channels = range(1)

        # data augmentation
        self.da_mirror_axes = (2, 3, 4)

        # validation
        self.val_use_DO = False
        self.val_use_train_mode = False # for dropout sampling
        self.val_num_repeats = 1 # only useful if dropout sampling
        self.val_batch_size = 1 # only useful if dropout sampling
        self.val_save_npz = True
        self.val_do_mirroring = True # test time data augmentation via mirroring
        self.val_write_images = True
        self.net_input_must_be_divisible_by = 16  # we could make a network class that has this as a property
        self.val_min_size = self.INPUT_PATCH_SIZE
        self.val_fn = None

        # CAREFUL! THIS IS A HACK TO MAKE PYTORCH 0.3 STATE DICTS COMPATIBLE WITH PYTORCH 0.4 (setting keep_runnings_
        # stats=True but not using them in validation. keep_runnings_stats was True before 0.3 but unused and defaults
        # to false in 0.4)
        self.val_use_moving_averages = False

    def get_network(self, train=True, pretrained_weights=None):
        net = Network(self.num_classes, len(self.selected_data_channels), self.net_base_num_layers,
                               self.net_dropout_p, softmax_helper, self.net_leaky_relu_slope, self.net_conv_use_bias,
                               self.net_norm_use_affine, True, self.net_do_DS)

        if pretrained_weights is not None:
            net.load_state_dict(
                torch.load(pretrained_weights, map_location=lambda storage, loc: storage))

        if train:
            net.train(True)
        else:
            net.train(False)
            net.apply(SetNetworkToVal(self.val_use_DO, self.val_use_moving_averages))
            net.do_ds = False

        optimizer = None
        self.lr_scheduler = None
        return net, optimizer

    def get_data_generators(self, fold):
        pass

    def get_split(self, fold, random_state=12345):
        pass

    def get_basic_generators(self, fold):
        pass

    def on_epoch_end(self, epoch):
        pass

    def preprocess(self, data):
        data = np.copy(data)
        for c in range(data.shape[0]):
            data[c] -= data[c].mean()
            data[c] /= data[c].std()
        return data

""" NETWORDK ARCHITECTURE"""
class EncodingModule(nn.Module):
    def __init__(self, in_channels, out_channels, filter_size=3, dropout_p=0.3, leakiness=1e-2, conv_bias=True,
                 inst_norm_affine=True, lrelu_inplace=True):
        nn.Module.__init__(self)
        self.dropout_p = dropout_p
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.bn_1 = nn.InstanceNorm3d(in_channels, affine=self.inst_norm_affine, track_running_stats=True)
        self.conv1 = nn.Conv3d(in_channels, out_channels, filter_size, 1, (filter_size - 1) // 2, bias=self.conv_bias)
        self.dropout = nn.Dropout3d(dropout_p)
        self.bn_2 = nn.InstanceNorm3d(in_channels, affine=self.inst_norm_affine, track_running_stats=True)
        self.conv2 = nn.Conv3d(out_channels, out_channels, filter_size, 1, (filter_size - 1) // 2, bias=self.conv_bias)

    def forward(self, x):
        skip = x
        x = F.leaky_relu(self.bn_1(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = self.conv1(x)
        if self.dropout_p is not None and self.dropout_p > 0:
            x = self.dropout(x)
        x = F.leaky_relu(self.bn_2(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = self.conv2(x)
        x = x + skip
        return x

class Upsample(nn.Module):
    def __init__(self, size=None, scale_factor=None, mode='nearest', align_corners=True):
        super(Upsample, self).__init__()
        self.align_corners = align_corners
        self.mode = mode
        self.scale_factor = scale_factor
        self.size = size

    def forward(self, x):
        return nn.functional.interpolate(x, size=self.size, scale_factor=self.scale_factor, mode=self.mode,
                                         align_corners=self.align_corners)

class LocalizationModule(nn.Module):
    def __init__(self, in_channels, out_channels, leakiness=1e-2, conv_bias=True, inst_norm_affine=True,
                 lrelu_inplace=True):
        nn.Module.__init__(self)
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.conv1 = nn.Conv3d(in_channels, in_channels, 3, 1, 1, bias=self.conv_bias)
        self.bn_1 = nn.InstanceNorm3d(in_channels, affine=self.inst_norm_affine, track_running_stats=True)
        self.conv2 = nn.Conv3d(in_channels, out_channels, 1, 1, 0, bias=self.conv_bias)
        self.bn_2 = nn.InstanceNorm3d(out_channels, affine=self.inst_norm_affine, track_running_stats=True)

    def forward(self, x):
        x = F.leaky_relu(self.bn_1(self.conv1(x)), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = F.leaky_relu(self.bn_2(self.conv2(x)), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        return x

class UpsamplingModule(nn.Module):
    def __init__(self, in_channels, out_channels, leakiness=1e-2, conv_bias=True, inst_norm_affine=True,
                 lrelu_inplace=True):
        nn.Module.__init__(self)
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.upsample = Upsample(scale_factor=2, mode="trilinear", align_corners=True)
        self.upsample_conv = nn.Conv3d(in_channels, out_channels, 3, 1, 1, bias=self.conv_bias)
        self.bn = nn.InstanceNorm3d(out_channels, affine=self.inst_norm_affine, track_running_stats=True)

    def forward(self, x):
        x = F.leaky_relu(self.bn(self.upsample_conv(self.upsample(x))), negative_slope=self.leakiness,
                         inplace=self.lrelu_inplace)
        return x

class DownsamplingModule(nn.Module):
    def __init__(self, in_channels, out_channels, leakiness=1e-2, conv_bias=True, inst_norm_affine=True,
                 lrelu_inplace=True):
        nn.Module.__init__(self)
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.bn = nn.InstanceNorm3d(in_channels, affine=self.inst_norm_affine, track_running_stats=True)
        self.downsample = nn.Conv3d(in_channels, out_channels, 3, 2, 1, bias=self.conv_bias)

    def forward(self, x):
        x = F.leaky_relu(self.bn(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        b = self.downsample(x)
        return x, b

class Network(nn.Module):
    def __init__(self, num_classes=4, num_input_channels=4, base_filters=16, dropout_p=0.3,
                 final_nonlin=softmax_helper, leakiness=1e-2, conv_bias=True, inst_norm_affine=True,
                 lrelu_inplace=True, do_ds=True):
        super(Network, self).__init__()

        self.do_ds = do_ds
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.conv_bias = conv_bias
        self.leakiness = leakiness
        self.final_nonlin = final_nonlin
        self.init_conv = nn.Conv3d(num_input_channels, base_filters, 3, 1, 1, bias=self.conv_bias)

        self.context1 = EncodingModule(base_filters, base_filters, 3, dropout_p, leakiness=1e-2, conv_bias=True,
                                       inst_norm_affine=True, lrelu_inplace=True)
        self.down1 = DownsamplingModule(base_filters, base_filters * 2, leakiness=1e-2, conv_bias=True,
                                        inst_norm_affine=True, lrelu_inplace=True)

        self.context2 = EncodingModule(2 * base_filters, 2 * base_filters, 3, dropout_p, leakiness=1e-2, conv_bias=True,
                                       inst_norm_affine=True, lrelu_inplace=True)
        self.down2 = DownsamplingModule(2 * base_filters, base_filters * 4, leakiness=1e-2, conv_bias=True,
                                        inst_norm_affine=True, lrelu_inplace=True)

        self.context3 = EncodingModule(4 * base_filters, 4 * base_filters, 3, dropout_p, leakiness=1e-2, conv_bias=True,
                                       inst_norm_affine=True, lrelu_inplace=True)
        self.down3 = DownsamplingModule(4 * base_filters, base_filters * 8, leakiness=1e-2, conv_bias=True,
                                        inst_norm_affine=True, lrelu_inplace=True)

        self.context4 = EncodingModule(8 * base_filters, 8 * base_filters, 3, dropout_p, leakiness=1e-2, conv_bias=True,
                                       inst_norm_affine=True, lrelu_inplace=True)
        self.down4 = DownsamplingModule(8 * base_filters, base_filters * 16, leakiness=1e-2, conv_bias=True,
                                        inst_norm_affine=True, lrelu_inplace=True)

        self.context5 = EncodingModule(16 * base_filters, 16 * base_filters, 3, dropout_p, leakiness=1e-2,
                                       conv_bias=True, inst_norm_affine=True, lrelu_inplace=True)

        self.bn_after_context5 = nn.InstanceNorm3d(16 * base_filters, affine=self.inst_norm_affine, track_running_stats=True)
        self.up1 = UpsamplingModule(16 * base_filters, 8 * base_filters, leakiness=1e-2, conv_bias=True,
                                    inst_norm_affine=True, lrelu_inplace=True)

        self.loc1 = LocalizationModule(16 * base_filters, 8 * base_filters, leakiness=1e-2, conv_bias=True,
                                       inst_norm_affine=True, lrelu_inplace=True)
        self.up2 = UpsamplingModule(8 * base_filters, 4 * base_filters, leakiness=1e-2, conv_bias=True,
                                    inst_norm_affine=True, lrelu_inplace=True)

        self.loc2 = LocalizationModule(8 * base_filters, 4 * base_filters, leakiness=1e-2, conv_bias=True,
                                       inst_norm_affine=True, lrelu_inplace=True)
        self.loc2_seg = nn.Conv3d(4 * base_filters, num_classes, 1, 1, 0, bias=False)
        self.up3 = UpsamplingModule(4 * base_filters, 2 * base_filters, leakiness=1e-2, conv_bias=True,
                                    inst_norm_affine=True, lrelu_inplace=True)

        self.loc3 = LocalizationModule(4 * base_filters, 2 * base_filters, leakiness=1e-2, conv_bias=True,
                                       inst_norm_affine=True, lrelu_inplace=True)
        self.loc3_seg = nn.Conv3d(2 * base_filters, num_classes, 1, 1, 0, bias=False)
        self.up4 = UpsamplingModule(2 * base_filters, 1 * base_filters, leakiness=1e-2, conv_bias=True,
                                    inst_norm_affine=True, lrelu_inplace=True)

        self.end_conv_1 = nn.Conv3d(2 * base_filters, 2 * base_filters, 3, 1, 1, bias=self.conv_bias)
        self.end_conv_1_bn = nn.InstanceNorm3d(2 * base_filters, affine=self.inst_norm_affine, track_running_stats=True)
        self.end_conv_2 = nn.Conv3d(2 * base_filters, 2 * base_filters, 3, 1, 1, bias=self.conv_bias)
        self.end_conv_2_bn = nn.InstanceNorm3d(2 * base_filters, affine=self.inst_norm_affine, track_running_stats=True)
        self.seg_layer = nn.Conv3d(2 * base_filters, num_classes, 1, 1, 0, bias=False)

    def forward(self, x):
        seg_outputs = []

        x = self.init_conv(x)
        x = self.context1(x)

        skip1, x = self.down1(x)
        x = self.context2(x)

        skip2, x = self.down2(x)
        x = self.context3(x)

        skip3, x = self.down3(x)
        x = self.context4(x)

        skip4, x = self.down4(x)
        x = self.context5(x)

        x = F.leaky_relu(self.bn_after_context5(x), negative_slope=self.leakiness, inplace=self.lrelu_inplace)
        x = self.up1(x)

        x = torch.cat((skip4, x), dim=1)
        x = self.loc1(x)
        x = self.up2(x)

        x = torch.cat((skip3, x), dim=1)
        x = self.loc2(x)
        loc2_seg = self.final_nonlin(self.loc2_seg(x))
        seg_outputs.append(loc2_seg)
        x = self.up3(x)

        x = torch.cat((skip2, x), dim=1)
        x = self.loc3(x)
        loc3_seg = self.final_nonlin(self.loc3_seg(x))
        seg_outputs.append(loc3_seg)
        x = self.up4(x)

        x = torch.cat((skip1, x), dim=1)
        x = F.leaky_relu(self.end_conv_1_bn(self.end_conv_1(x)), negative_slope=self.leakiness,
                         inplace=self.lrelu_inplace)
        x = F.leaky_relu(self.end_conv_2_bn(self.end_conv_2(x)), negative_slope=self.leakiness,
                         inplace=self.lrelu_inplace)
        x = self.final_nonlin(self.seg_layer(x))
        seg_outputs.append(x)

        if self.do_ds:
            return seg_outputs[::-1]
        else:
            return seg_outputs[-1]

"""DATA LOADING"""
def resize_image(image, old_spacing, new_spacing, order=3):
    new_shape = (int(np.round(old_spacing[0]/new_spacing[0]*float(image.shape[0]))),
                 int(np.round(old_spacing[1]/new_spacing[1]*float(image.shape[1]))),
                 int(np.round(old_spacing[2]/new_spacing[2]*float(image.shape[2]))))
    return resize(image, new_shape, order=order, mode='edge', cval=0, anti_aliasing=False)

def preprocess_image(itk_image, is_seg=False, spacing_target=(1, 0.5, 0.5)):
    spacing = np.array(itk_image.GetSpacing())[[2, 1, 0]]
    image = sitk.GetArrayFromImage(itk_image).astype(float)

    if len(image.shape) == 4:
        image = np.squeeze(image, axis=0)

    if not is_seg:
        if np.any([[i != j] for i, j in zip(spacing, spacing_target)]):
            image = resize_image(image, spacing, spacing_target).astype(np.float32)

        image -= image.mean()
        image /= image.std()
    else:
        new_shape = (int(np.round(spacing[0] / spacing_target[0] * float(image.shape[0]))),
                     int(np.round(spacing[1] / spacing_target[1] * float(image.shape[1]))),
                     int(np.round(spacing[2] / spacing_target[2] * float(image.shape[2]))))
        image = resize_segmentation(image, new_shape, 1)
    return image

def load_and_preprocess(mri_file):
    images = {}
    # t1
    images["T1"] = mri_file

    properties_dict = {
        "spacing": images["T1"].GetSpacing(),
        "direction": images["T1"].GetDirection(),
        "size": images["T1"].GetSize(),
        "origin": images["T1"].GetOrigin()
    }

    direction = list(properties_dict["direction"])
    if len(direction) == 16: 
        direction = direction[:11]
        direction.pop(3)
        direction.pop(6)
        properties_dict["direction"] = tuple(direction)

    for k in images.keys():
        images[k] = preprocess_image(images[k], is_seg=False, spacing_target=(1.5, 1.5, 1.5))

    properties_dict['size_before_cropping'] = images["T1"].shape
    imgs = []
    for seq in ['T1']:
        imgs.append(images[seq][None])
    all_data = np.vstack(imgs)
    return all_data, properties_dict

def segmentation_nifti(segmentation, dct, order=1):
    '''
    segmentation must have the same spacing as the original nifti (for now). segmentation may have been cropped out
    of the original image

    dct:
    size_before_cropping
    brain_bbox
    size -> this is the original size of the dataset, if the image was not resampled, this is the same as size_before_cropping
    spacing
    origin
    direction

    :param segmentation:
    :param dct:
    :return:
    '''
    old_size = dct.get('size_before_cropping')
    bbox = dct.get('brain_bbox')
    if bbox is not None:
        seg_old_size = np.zeros(old_size)
        for c in range(3):
            bbox[c][1] = np.min((bbox[c][0] + segmentation.shape[c], old_size[c]))
        seg_old_size[bbox[0][0]:bbox[0][1],
                     bbox[1][0]:bbox[1][1],
                     bbox[2][0]:bbox[2][1]] = segmentation
    else:
        seg_old_size = segmentation
    # if np.any(np.array(seg_old_size) != np.array(dct['size'])[[2, 1, 0]]):
    seg_old_spacing = resize_segmentation(seg_old_size, np.array(dct['size'])[[2, 1, 0]], order=order)
    # else:
        # seg_old_spacing = seg_old_size
    seg_resized_itk = sitk.GetImageFromArray(seg_old_spacing.astype(np.int32))
    seg_resized_itk.SetSpacing(np.array(dct['spacing'])[[0, 1, 2]])
    seg_resized_itk.SetOrigin(dct['origin'])
    seg_resized_itk.SetDirection(dct['direction'])
    return seg_resized_itk

def resize_segmentation(segmentation, new_shape, order=3, cval=0):
    '''
    Taken from batchgenerators (https://github.com/MIC-DKFZ/batchgenerators) to prevent dependency

    Resizes a segmentation map. Supports all orders (see skimage documentation). Will transform segmentation map to one
    hot encoding which is resized and transformed back to a segmentation map.
    This prevents interpolation artifacts ([0, 0, 2] -> [0, 1, 2])
    :param segmentation:
    :param new_shape:
    :param order:
    :return:
    '''
    tpe = segmentation.dtype
    unique_labels = np.unique(segmentation)
    assert len(segmentation.shape) == len(new_shape), "new shape must have same dimensionality as segmentation"
    if order == 0:
        return resize(segmentation, new_shape, order, mode="constant", cval=cval, clip=True, anti_aliasing=False).astype(tpe)
    else:
        reshaped = np.zeros(new_shape, dtype=segmentation.dtype)

        for i, c in enumerate(unique_labels):
            reshaped_multihot = resize((segmentation == c).astype(float), new_shape, order, mode="edge", clip=True, anti_aliasing=False)
            reshaped[reshaped_multihot >= 0.5] = c
        return reshaped

""" PREDICT CASE """
def pad_patient_3D(patient, shape_must_be_divisible_by=16, min_size=None):
    if not (isinstance(shape_must_be_divisible_by, list) or isinstance(shape_must_be_divisible_by, tuple)):
        shape_must_be_divisible_by = [shape_must_be_divisible_by] * 3
    shp = patient.shape
    new_shp = [shp[0] + shape_must_be_divisible_by[0] - shp[0] % shape_must_be_divisible_by[0],
               shp[1] + shape_must_be_divisible_by[1] - shp[1] % shape_must_be_divisible_by[1],
               shp[2] + shape_must_be_divisible_by[2] - shp[2] % shape_must_be_divisible_by[2]]
    for i in range(len(shp)):
        if shp[i] % shape_must_be_divisible_by[i] == 0:
            new_shp[i] -= shape_must_be_divisible_by[i]
    if min_size is not None:
        new_shp = np.max(np.vstack((np.array(new_shp), np.array(min_size))), 0)
    return reshape_by_padding_upper_coords(patient, new_shp, 0), shp

def reshape_by_padding_upper_coords(image, new_shape, pad_value=None):
    shape = tuple(list(image.shape))
    new_shape = tuple(np.max(np.concatenate((shape, new_shape)).reshape((2,len(shape))), axis=0))
    if pad_value is None:
        if len(shape) == 2:
            pad_value = image[0,0]
        elif len(shape) == 3:
            pad_value = image[0, 0, 0]
        else:
            raise ValueError("Image must be either 2 or 3 dimensional")
    res = np.ones(list(new_shape), dtype=image.dtype) * pad_value
    if len(shape) == 2:
        res[0:0+int(shape[0]), 0:0+int(shape[1])] = image
    elif len(shape) == 3:
        res[0:0+int(shape[0]), 0:0+int(shape[1]), 0:0+int(shape[2])] = image
    return res

def predict_case_3D_net(net, patient_data, do_mirroring, num_repeats, BATCH_SIZE=None,
                           new_shape_must_be_divisible_by=16, min_size=None, main_device=0, mirror_axes=(2, 3, 4)):
    with torch.no_grad():
        pad_res = []
        for i in range(patient_data.shape[0]):
            t, old_shape = pad_patient_3D(patient_data[i], new_shape_must_be_divisible_by, min_size)
            pad_res.append(t[None])

        patient_data = np.vstack(pad_res)

        new_shp = patient_data.shape

        data = np.zeros(tuple([1] + list(new_shp)), dtype=np.float32)

        data[0] = patient_data

        if BATCH_SIZE is not None:
            data = np.vstack([data] * BATCH_SIZE)

        a = torch.rand(data.shape).float()

        if main_device == 'cpu':
            pass
        else:
            a = a.cuda(main_device)

        if do_mirroring:
            x = 8
        else:
            x = 1
        all_preds = []
        for i in range(num_repeats):
            for m in range(x):
                data_for_net = np.array(data)
                do_stuff = False
                if m == 0:
                    do_stuff = True
                    pass
                if m == 1 and (4 in mirror_axes):
                    do_stuff = True
                    data_for_net = data_for_net[:, :, :, :, ::-1]
                if m == 2 and (3 in mirror_axes):
                    do_stuff = True
                    data_for_net = data_for_net[:, :, :, ::-1, :]
                if m == 3 and (4 in mirror_axes) and (3 in mirror_axes):
                    do_stuff = True
                    data_for_net = data_for_net[:, :, :, ::-1, ::-1]
                if m == 4 and (2 in mirror_axes):
                    do_stuff = True
                    data_for_net = data_for_net[:, :, ::-1, :, :]
                if m == 5 and (2 in mirror_axes) and (4 in mirror_axes):
                    do_stuff = True
                    data_for_net = data_for_net[:, :, ::-1, :, ::-1]
                if m == 6 and (2 in mirror_axes) and (3 in mirror_axes):
                    do_stuff = True
                    data_for_net = data_for_net[:, :, ::-1, ::-1, :]
                if m == 7 and (2 in mirror_axes) and (3 in mirror_axes) and (4 in mirror_axes):
                    do_stuff = True
                    data_for_net = data_for_net[:, :, ::-1, ::-1, ::-1]

                if do_stuff:
                    _ = a.data.copy_(torch.from_numpy(np.copy(data_for_net)))
                    p = net(a)  # np.copy is necessary because ::-1 creates just a view i think
                    p = p.data.cpu().numpy()

                    if m == 0:
                        pass
                    if m == 1 and (4 in mirror_axes):
                        p = p[:, :, :, :, ::-1]
                    if m == 2 and (3 in mirror_axes):
                        p = p[:, :, :, ::-1, :]
                    if m == 3 and (4 in mirror_axes) and (3 in mirror_axes):
                        p = p[:, :, :, ::-1, ::-1]
                    if m == 4 and (2 in mirror_axes):
                        p = p[:, :, ::-1, :, :]
                    if m == 5 and (2 in mirror_axes) and (4 in mirror_axes):
                        p = p[:, :, ::-1, :, ::-1]
                    if m == 6 and (2 in mirror_axes) and (3 in mirror_axes):
                        p = p[:, :, ::-1, ::-1, :]
                    if m == 7 and (2 in mirror_axes) and (3 in mirror_axes) and (4 in mirror_axes):
                        p = p[:, :, ::-1, ::-1, ::-1]
                    all_preds.append(p)

        stacked = np.vstack(all_preds)[:, :, :old_shape[0], :old_shape[1], :old_shape[2]]
        predicted_segmentation = stacked.mean(0).argmax(0)
        uncertainty = stacked.var(0)
        bayesian_predictions = stacked
        softmax_pred = stacked.mean(0)
    return predicted_segmentation, bayesian_predictions, softmax_pred, uncertainty

""" RUN """
def apply_bet(img, bet):
    img_itk = sitk.GetArrayFromImage(img)
    img_bet = sitk.GetArrayFromImage(bet)
    img_itk[img_bet == 0] = 0
    return img_itk

def run_hd_bet(input_data, model_path, mode="accurate", device=0,
               postprocess=False, do_tta=True, keep_mask=True, overwrite=True):
    """

    :param input_data: str or list/tuple of str
    :param mode: fast or accurate
    :param config_file: config.py
    :param device: either int (for device id) or 'cpu'
    :param postprocess: whether to do postprocessing or not. Postprocessing here consists of simply discarding all
    but the largest predicted connected component. Default False
    :param do_tta: whether to do test time data augmentation by mirroring along all axes. Default: True. If you use
    CPU you may want to turn that off to speed things up
    :return:
    """

    list_of_param_files = []

    if mode == 'fast':
        params_file = get_params_fname(model_path, 0)

        list_of_param_files.append(params_file)
    elif mode == 'accurate':
        for i in range(5):
            params_file = get_params_fname(model_path, i)
            list_of_param_files.append(params_file)
    else:
        raise ValueError("Unknown value for mode: %s. Expected: fast or accurate" % mode)

    assert all([os.path.isfile(i) for i in list_of_param_files]), "Could not find parameter files"

    cf = HD_BET_Config()

    net, _ = cf.get_network(cf.val_use_train_mode, None)
    if device == "cpu":
        net = net.cpu()
    else:
        net.cuda(device)

    params = []
    for p in list_of_param_files:
        params.append(torch.load(p, map_location=lambda storage, loc: storage))


        try:
            data, data_dict = load_and_preprocess(input_data)
        except RuntimeError:
            print("\nERROR\nCould not read file", input_data, "\n")
            continue
        except AssertionError as e:
            print(e)
            continue


        softmax_preds = []

        for i, p in enumerate(params):
            net.load_state_dict(p)
            net.eval()
            net.apply(SetNetworkToVal(False, False))
            _, _, softmax_pred, _ = predict_case_3D_net(net, data, do_tta, cf.val_num_repeats,
                                                        cf.val_batch_size, cf.net_input_must_be_divisible_by,
                                                        cf.val_min_size, device, cf.da_mirror_axes)
            softmax_preds.append(softmax_pred[None])

        seg = np.argmax(np.vstack(softmax_preds).mean(0), 0)

        if postprocess:
            seg = postprocess_prediction(seg)

        seg = segmentation_nifti(seg, data_dict)
        final = apply_bet(input_data, seg)
        return final, sitk.GetArrayFromImage(seg)

if __name__ == "__main__":
    path = r"F:\Dev\taima-2022\data\nifti\AX_T1_MPRAGE.nii.gz"
    run_hd_bet(sitk.ReadImage(path), model_path=r"C:\Users\467355\Documents\HD-BET-master\HD_BET\hd-bet_params")


WINDOW_WIDTH = 1800
WINDOW_HEIGHT = 960 

ONED_WIDTH = 900
ONED_HEIGHT = 900

THREED_WIDTH = 500
THREED_HEIGHT = 500

INFO_WIDTH = 400

SCROLL_WIDTH = 260
THUMBNAIL_WIDTH = 200

BG_COLOR = "#0B1117 "
LIGHT_BG_COLOR = "#131B2A"
VERY_LIGHT_BG_COLOR = "#507BA6"
BORDER_COLOR = "#243242"
LIGHT_BORDER_COLOR = "#374558"
TEXT_COLOR = "#E0E1DD"




import sys
import os
import nibabel as nib
import numpy as np
import matplotlib.pyplot as plt
from PySide6.QtWidgets import QFrame, QHeaderView, QApplication, QPushButton, QSlider, QDialog, QComboBox, QToolButton, QLineEdit, QLabel, QMainWindow, QCheckBox, QScrollArea, QStackedLayout, QVBoxLayout, QHBoxLayout, QTableWidget, QTableWidgetItem, QWidget, QFileDialog, QMenu
from PySide6.QtGui import QImage, QPixmap, QAction, QGuiApplication, QIcon, QDragEnterEvent, QDropEvent
from PySide6.QtCore import Qt, QEvent, QRect, QSize, QSettings, QUrl, QThreadPool, QRunnable, Signal, Slot, QObject, QEventLoop, QWaitCondition, QMutex
import pyqtgraph as pg
import matplotlib.pyplot as plt
import glob
from io import StringIO
import imageio
import tempfile
import threading

from pydactim.transformation import (n4_bias_field_correction,
    registration, apply_transformation, resample, histogram_matching, 
    apply_mask, substract, susan, normalize, crop, apply_crop, copy_affine, skull_stripping, 
    remove_small_object, tissue_classifier, add_tissue_class, prediction_glioma, prediction_multiple_sclerosis, uncertainty_prediction_glioma)

from pydactim.sorting import sort_dicom
from pydactim.conversion import convert_dicom_to_nifti
from .custom import ThumbnailFrame, ScrollArea, ComboBox, AnimatedToggle
from .view_panel import ViewPanel
from .utils import create_thumbnail, get_darkModePalette, reset_layout
from .settings import *

class AiModelPathDialog(QDialog):
    def __init__(self, settings):
        super().__init__()

        self.settings = settings

        self.setWindowTitle("Select AI Model Directories")
        self.layout = QVBoxLayout()

        self.skull_stripping_label = QLabel("Skull Stripping Model Directory:")
        self.layout.addWidget(self.skull_stripping_label)
        self.skull_stripping_path = QLabel(settings.value("ai_model_paths/skull_stripping", ""))
        self.layout.addWidget(self.skull_stripping_path)
        self.skull_stripping_button = QPushButton("Select Skull Stripping Model Directory")
        self.skull_stripping_button.clicked.connect(self.select_skull_stripping_path)
        self.layout.addWidget(self.skull_stripping_button)

        self.add_separator()

        self.glioma_segmentation_label = QLabel("Glioma Segmentation Model directory (with weights and landmarks):")
        self.layout.addWidget(self.glioma_segmentation_label)
        self.glioma_segmentation_path = QLabel(settings.value("ai_model_paths/glioma", ""))
        self.layout.addWidget(self.glioma_segmentation_path)
        self.glioma_segmentation_button = QPushButton("Select Glioma Segmentation Model Directory")
        self.glioma_segmentation_button.clicked.connect(self.select_glioma_segmentation_path)
        self.layout.addWidget(self.glioma_segmentation_button)

        self.add_separator()

        self.ms_segmentation_label = QLabel("MS Segmentation Model directory (with weights and landmarks):")
        self.layout.addWidget(self.ms_segmentation_label)
        self.ms_segmentation_path = QLabel(settings.value("ai_model_paths/ms", ""))
        self.layout.addWidget(self.ms_segmentation_path)
        self.ms_segmentation_button = QPushButton("Select MS Segmentation Model Directory")
        self.ms_segmentation_button.clicked.connect(self.select_ms_segmentation_path)
        self.layout.addWidget(self.ms_segmentation_button)

        self.add_separator()

        self.save_button = QPushButton("Save")
        self.save_button.clicked.connect(self.save_paths)
        self.layout.addWidget(self.save_button)

        self.setLayout(self.layout)

    def add_separator(self):
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setFrameShadow(QFrame.Sunken)
        self.layout.addWidget(separator)

    def select_skull_stripping_path(self):
        path = QFileDialog.getExistingDirectory(self, "Select Skull Stripping Model Directory")
        if path: self.skull_stripping_path.setText(path)

    def select_glioma_segmentation_path(self):
        path = QFileDialog.getExistingDirectory(self, "Select Glioma Segmentation Model Directory")
        if path: self.glioma_segmentation_path.setText(path)

    def select_ms_segmentation_path(self):
        path = QFileDialog.getExistingDirectory(self, "Select MS Segmentation Model Directory")
        if path: self.ms_segmentation_path.setText(path)

    def save_paths(self):
        self.settings.setValue("ai_model_paths/skull_stripping", self.skull_stripping_path.text())
        self.settings.setValue("ai_model_paths/glioma", self.glioma_segmentation_path.text())
        self.settings.setValue("ai_model_paths/ms", self.ms_segmentation_path.text())
        # Save other AI model paths similarly
        self.accept()

class NiftiViewer(QMainWindow):
    def __init__(self):
        super().__init__()
        print("INFO - Initializing Pydactim Viewer")
        self.setWindowTitle("PyDactim Viewer")

        # self.nifti_found = glob.glob('./**/*.nii.gz', recursive=True)
        self.init_tabs()
        # self.init_all()
        # self.open_volume()
        self.dynamic_windows = []

        settings = self.init_config()
        last_path = settings.value(f"recent_paths/recent_path_1", "")
        print(f"{last_path = }")
        if os.path.exists(last_path):
            if os.path.isdir(last_path): self.open_volume(last_path)
            else: self.add_file(last_path)

    def init_all(self):
        self.filename = self.nifti_found[0]

        print("INFO - Creating UI")
        self.init_ui()
        # print("INFO - Loading images")
        # self.load_sequences(self.nifti_found)

        self.lut = None
        self.overlay_lut = None

        self.overlay = False
        self.overlay_path = None

        # sys.stdout = CustomOutputStream(self.log_label, self.log_scroll)

    def init_tabs(self):
        print("INFO - Creating menu tab")
        # Add a menu bar
        menubar = self.menuBar()
        self.file_menu = menubar.addMenu("File")
        edit_menu = menubar.addMenu("Edit")
        image_menu = menubar.addMenu("Image")
        transform_menu = menubar.addMenu("Scripts")
        view_menu = menubar.addMenu("View")

        # Add actions to the File menu
        action = QAction("Open directory", self)
        action.triggered.connect(lambda x: self.open_volume(None))
        self.file_menu.addAction(action)

        action = QAction("Add file", self)
        action.triggered.connect(lambda x: self.add_file(None))
        self.file_menu.addAction(action)

        self.recent_submenu = self.file_menu.addMenu("Recent data")
        settings = self.init_config()
        recent_paths = [settings.value(f"recent_paths/recent_path_{i}", "") for i in range(1, 11)]

        for path in recent_paths:
            if os.path.exists(path):
                action = QAction(path, self)
                if os.path.isdir(path): action.triggered.connect(lambda checked, p=path: self.open_volume(p))
                else: action.triggered.connect(lambda checked, p=path: self.add_file(p))
                self.recent_submenu.addAction(action)

        action = QAction("Settings", self)
        action.triggered.connect(self.settings)
        self.file_menu.addAction(action)

        action = QAction("Screenshot", self)
        action.triggered.connect(self.screenshot)
        self.file_menu.addAction(action)

        action = QAction("Gif", self)
        action.triggered.connect(self.gifshot)
        self.file_menu.addAction(action)

        action = QAction("Close", self)
        action.triggered.connect(lambda x: sys.exit(1))
        self.file_menu.addAction(action)

        action = QAction("Sort DICOM directory", self)
        action.triggered.connect(lambda x: self.transforms("Sort DICOM"))
        edit_menu.addAction(action)

        action = QAction("Convert DICOM to NIfTI", self)
        action.triggered.connect(lambda x: self.transforms("Convert DICOM"))
        edit_menu.addAction(action)

        # Add actions to the Tool menu
        action = QAction("Histogram whole image", self)
        action.triggered.connect(self.whole_histogram)
        image_menu.addAction(action)

        action = QAction("Histogram current image", self)
        action.triggered.connect(self.add_current_histogram)
        image_menu.addAction(action)

        # action = QAction("Reset image", self)
        # action.triggered.connect(self.reset_image)
        # image_menu.addAction(action)

        action = QAction("Reset windowing", self)
        action.triggered.connect(self.reset_contrast)
        image_menu.addAction(action)

        # Add actions to the View menu
        action = QAction("Axial view", self)
        action.triggered.connect(lambda x: self.toggle_view("Axial"))
        view_menu.addAction(action)

        action = QAction("Coronal view", self)
        action.triggered.connect(lambda x: self.toggle_view("Coronal"))
        view_menu.addAction(action)

        action = QAction("Sagittal view", self)
        action.triggered.connect(lambda x: self.toggle_view("Sagittal"))
        view_menu.addAction(action)

        action = QAction("Mosaic view", self)
        action.triggered.connect(lambda x: self.toggle_view("Mosaic"))
        view_menu.addAction(action)

        action = QAction("3D view", self)
        action.triggered.connect(lambda x: self.toggle_view("3D"))
        view_menu.addAction(action)

        # AI
        ai_submenu = transform_menu.addMenu("AI segmentation")

        action = QAction("Brain", self)
        action.triggered.connect(lambda x: self.transforms("Skull stripping"))
        ai_submenu.addAction(action)

        ai_action_1 = QAction("Glioma", self)
        ai_action_1.triggered.connect(lambda x: self.transforms("Glioma"))
        ai_submenu.addAction(ai_action_1)

        ai_action_3 = QAction("Meningioma", self)
        ai_action_3.triggered.connect(lambda x: self.transforms("Meningioma"))
        ai_submenu.addAction(ai_action_3)

        ai_action_2 = QAction("Multiple Sclerosis", self)
        ai_action_2.triggered.connect(lambda x: self.transforms("Multiple Sclerosis"))
        ai_submenu.addAction(ai_action_2)

        # Crop
        crop_submenu = transform_menu.addMenu("Crop")

        action = QAction("Auto crop", self)
        action.triggered.connect(lambda x: self.transforms("Auto crop"))
        crop_submenu.addAction(action)

        action = QAction("Apply crop", self)
        action.triggered.connect(lambda x: self.transforms("Apply crop"))
        crop_submenu.addAction(action)

        filter_submenu = transform_menu.addMenu("Filter")

        action = QAction("Remove islands", self)
        action.triggered.connect(lambda x: self.transforms("Remove islands"))
        filter_submenu.addAction(action)

        action = QAction("Susan", self)
        action.triggered.connect(lambda x: self.transforms("Susan"))
        filter_submenu.addAction(action)

        registration_submenu = transform_menu.addMenu("Registration")

        action = QAction("Registration", self)
        action.triggered.connect(lambda x: self.transforms("Registration"))
        registration_submenu.addAction(action)

        action = QAction("Apply registration", self)
        action.triggered.connect(lambda x: self.transforms("Apply registration"))
        registration_submenu.addAction(action)

        other_submenu = transform_menu.addMenu("Others")

        action = QAction("Apply mask", self)
        action.triggered.connect(lambda x: self.transforms("Apply mask"))
        other_submenu.addAction(action)

        action = QAction("Bias field correction", self)
        action.triggered.connect(lambda x: self.transforms("Bias field correction"))
        other_submenu.addAction(action)

        action = QAction("Resample voxels", self)
        action.triggered.connect(lambda x: self.transforms("Resample voxels"))
        other_submenu.addAction(action)

    def init_ui(self):
        print("INFO - Creating left panel")
        # Sequences layout
        self.scroll_area = ScrollArea()

        print("INFO - Creating right panel")

        # ICON GUI
        self.selected_widget = "info_widget"
        self.icon_widget = QWidget()
        self.icon_layout = QHBoxLayout()

        self.icon_info_button = QToolButton()
        self.icon_info_button.setIcon(QIcon("info.png"))
        self.icon_info_button.setText("Image details")
        self.icon_info_button.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.icon_info_button.setIconSize(QSize(50,50))
        self.icon_info_button.setCursor(Qt.PointingHandCursor)
        self.icon_info_button.clicked.connect(self.show_info)

        self.icon_perf_button = QToolButton()
        self.icon_perf_button.setIcon(QIcon("perf.png"))
        self.icon_perf_button.setText("Perfusion Maps")
        self.icon_perf_button.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.icon_perf_button.setIconSize(QSize(50,50))
        self.icon_perf_button.setCursor(Qt.PointingHandCursor)
        self.icon_perf_button.clicked.connect(self.show_perf)

        self.icon_layout.addWidget(self.icon_info_button)
        self.icon_layout.addWidget(self.icon_perf_button)
        self.icon_layout.addStretch()
        self.icon_widget.setLayout(self.icon_layout)

        # INFO GUI
        self.info_widget = QWidget()
        self.info_layout = QVBoxLayout()

        self.text_label = QLabel(alignment=Qt.AlignmentFlag.AlignTop)
        self.text_label.setStyleSheet(f"background-color: {LIGHT_BG_COLOR}")
        self.text_label.setStyleSheet("border: 1px solid white")
        self.text_label.setStyleSheet("padding: 0px 15px")
        self.text_label.setMinimumSize(INFO_WIDTH, 0)

        self.plot_aif = pg.PlotWidget()
        self.plot_aif.getPlotItem().hideAxis('left')

        self.lut_widget = QWidget()
        self.lut_layout = QVBoxLayout()
        self.lut_label = QLabel("Look-up-table")
        self.lut_button = ComboBox()
        self.lut_button.addItems(['Grayscale', 'Red', 'Green', 'Blue', 'Rainbow', 'Spectral', 'Flow', 'Viridis', 'Qualitative', 'Random'])
        self.lut_button.currentTextChanged.connect(self.change_lut)
        self.lut_layout.addWidget(self.lut_label)
        self.lut_layout.addWidget(self.lut_button)
        self.lut_widget.setLayout(self.lut_layout)

        self.log_title = QLabel("<h2><b>Log:</b></h2>")
        self.log_label = QLabel()
        self.log_label.setAlignment(Qt.AlignTop)
        self.log_scroll = QScrollArea()
        self.log_scroll.setStyleSheet(f"background-color: black; color: {TEXT_COLOR};")
        self.log_scroll.setWidget(self.log_label)
        self.log_scroll.setFixedHeight(300)
        self.log_scroll.setWidgetResizable(True)
        self.log_layout = QVBoxLayout()
        self.log_layout.addStretch()
        self.log_layout.addWidget(self.log_title)
        self.log_layout.addWidget(self.log_scroll)
        self.log_widget = QWidget()
        self.log_widget.setLayout(self.log_layout)

        self.info_layout.addWidget(self.text_label)
        self.info_layout.addWidget(self.lut_widget)
        self.info_layout.addStretch()
        self.info_layout.addWidget(self.log_widget)
        self.info_widget.setLayout(self.info_layout)

        # PERF GUI
        self.perf_widget = QWidget()
        self.perf_layout = QVBoxLayout()
        self.plot_aif = pg.PlotWidget()
        self.plot_aif.getPlotItem().hideAxis('left')

        self.roi_widget = QWidget()
        self.roi_layout = QHBoxLayout()

        self.roi_x_widget = QWidget()
        self.roi_x_layout = QHBoxLayout()
        self.roi_x_title = QLabel("ROI x")
        self.roi_x = QLineEdit()
        self.roi_x.setFixedWidth(50)
        self.roi_x_layout.addWidget(self.roi_x_title)
        self.roi_x_layout.addWidget(self.roi_x)
        self.roi_x_layout.setAlignment(Qt.AlignLeft)
        self.roi_x_widget.setLayout(self.roi_x_layout)

        self.roi_y_widget = QWidget()
        self.roi_y_layout = QHBoxLayout()
        self.roi_y_title = QLabel("ROI y")
        self.roi_y = QLineEdit()
        self.roi_y.setFixedWidth(50)
        self.roi_y_layout.addWidget(self.roi_y_title)
        self.roi_y_layout.addWidget(self.roi_y)
        self.roi_y_layout.setAlignment(Qt.AlignLeft)
        self.roi_y_widget.setLayout(self.roi_y_layout)

        self.roi_z_widget = QWidget()
        self.roi_z_layout = QHBoxLayout()
        self.roi_z_title = QLabel("ROI z")
        self.roi_z = QLineEdit()
        self.roi_z.setFixedWidth(50)
        self.roi_z_layout.addWidget(self.roi_z_title)
        self.roi_z_layout.addWidget(self.roi_z)
        self.roi_z_layout.setAlignment(Qt.AlignLeft)
        self.roi_z_widget.setLayout(self.roi_z_layout)

        self.roi_table = QTableWidget()
        roi_header = ["Mean", "Std", "Min", "Max", "mm3"]
        self.roi_table.setColumnCount(len(roi_header))
        self.roi_table.setHorizontalHeaderLabels(roi_header)
        self.roi_table.setRowCount(1)
        self.roi_table.verticalHeader().setVisible(False)
        _ = [self.roi_table.setColumnWidth(i,70) for i in range(len(roi_header))]
        self.roi_table.horizontalHeader().setStretchLastSection(True)


        self.roi_layout.addWidget(self.roi_x_widget)
        self.roi_layout.addWidget(self.roi_y_widget)
        self.roi_layout.addWidget(self.roi_z_widget)
        self.roi_widget.setLayout(self.roi_layout)

        self.perf_layout.addWidget(self.plot_aif)
        self.perf_layout.addWidget(self.roi_widget)
        self.perf_layout.addWidget(self.roi_table)
        self.perf_widget.setLayout(self.perf_layout)

        # RIGHT LAYOUT
        self.right_layout = QVBoxLayout()
        self.right_layout.addWidget(self.icon_widget)
        self.right_layout.addWidget(self.info_widget)
        self.right_layout.addWidget(self.perf_widget)
        self.right_layout.addStretch()
        self.perf_widget.hide()
 
        self.right_widget = QWidget()
        self.right_widget.setLayout(self.right_layout)

        print("INFO - Creating middle panel")
        # View label that contains the image label(s)
        self.view_widget = ViewPanel(self.filename, self.update_text, self.plot_aif, self.update_roi)
        self.view_widget.update_text()

        print("INFO - Creating main layout")
        self.main_layout = QHBoxLayout()
        self.main_layout.addWidget(self.scroll_area)
        self.main_layout.addWidget(self.view_widget)
        self.main_layout.addWidget(self.right_widget)

        central_widget = QWidget(self)
        central_widget.setLayout(self.main_layout)
        self.setCentralWidget(central_widget)

        print("INFO - Adding style sheet")
        # self.update_image()
        # self.create_histogram()
        self.setStyleSheet(
            """
            QWidget {
                background-color: $BG_COLOR;
            }
          
            QScrollArea {
                background-color: $BG_COLOR;
                border: 0px solid $BG_COLOR;
            }
            QScrollBar:vertical {
                border: 0px solid $BG_COLOR;
                background: $LIGHT_BG_COLOR;
                width:10px;    
                margin: 0px 0px 0px 0px;
            }
            QScrollBar::handle:vertical {         
        
                min-height: 0px;
                border: 0px solid red;
                border-radius: 4px;
                background-color: $BORDER_COLOR;
            }
            QScrollBar::add-line:vertical {       
                height: 0px;
                subcontrol-position: bottom;
                subcontrol-origin: margin;
            }
            QScrollBar::sub-line:vertical {
                height: 0 px;
                subcontrol-position: top;
                subcontrol-origin: margin;
            }
            QToolButton::hover{
                background-color: #000000;
            }
            """.replace('$BG_COLOR', BG_COLOR)\
            .replace('$LIGHT_BG_COLOR', LIGHT_BG_COLOR)\
            .replace('$BORDER_COLOR', BORDER_COLOR)
        )

    def init_config(self):
        temp_dir = tempfile.gettempdir()
        config_path = os.path.join(temp_dir, 'pydactim_view_config.ini')
        settings = QSettings(config_path, QSettings.IniFormat)

        for i in range(1, 11):
            if not settings.contains(f"recent_paths/recent_path_{i}"):
                settings.setValue(f"recent_paths/recent_path_{i}", "")

        if not settings.contains("ai_model_paths/skull_stripping"):
            settings.setValue("ai_model_paths/skull_stripping", "")

        if not settings.contains("ai_model_paths/glioma"):
            settings.setValue("ai_model_paths/glioma", "")

        if not settings.contains("ai_model_paths/ms"):
            settings.setValue("ai_model_paths/ms", "")

        return settings

    def settings(self):
        self.open_ai_model_path_dialog()

    def add_recent_path(self, path):
        settings = self.init_config()
        recent_paths = [settings.value(f"recent_paths/recent_path_{i}", "") for i in range(1, 11)]
        
        # Ensure the path is not already in the list
        if path in recent_paths:
            recent_paths.remove(path)
        
        recent_paths.insert(0, path)
        
        # Limit to 10 recent paths
        if len(recent_paths) > 10:
            recent_paths = recent_paths[:10]
        
        for i in range(1, 11):
            settings.setValue(f"recent_paths/recent_path_{i}", recent_paths[i-1] if i-1 < len(recent_paths) else "")

        self.recent_submenu.clear()
        for path in recent_paths:
            if os.path.exists(path):
                action = QAction(path, self)
                if os.path.isdir(path): action.triggered.connect(lambda checked, p=path: self.open_volume(p))
                else: action.triggered.connect(lambda checked, p=path: self.add_file(p))
                self.recent_submenu.addAction(action)

    def open_ai_model_path_dialog(self):
        settings = self.init_config()
        dialog = AiModelPathDialog(settings)
        dialog.exec_()

    def show_info(self):
        self.perf_widget.hide()
        self.info_widget.show()
        self.selected_widget = "info_widget"
        self.view_widget.axial_label.selected_widget = self.selected_widget
        self.view_widget.axial_label.update_image()

    def show_perf(self):
        self.info_widget.hide()
        self.perf_widget.show()
        self.selected_widget = "perf_widget"
        self.view_widget.axial_label.selected_widget = self.selected_widget
        self.view_widget.axial_label.update_image()

    def open_volume(self, path=None):
        if path is None: self.dirname = QFileDialog.getExistingDirectory(self, "Select a directory", "C:\\", QFileDialog.ShowDirsOnly)
        else: self.dirname = path
        # self.dirname = "D:/Studies/GLIOBIOPSY/data/sub-018/"
        self.nifti_found = glob.glob(os.path.join(self.dirname, '**/*.nii*'), recursive=True)
        if not hasattr(self, "thumbnail_title"): 
            self.init_all()
            self.load_sequences(self.nifti_found)
        else: self.load_sequences(self.nifti_found)
        self.add_recent_path(self.dirname)

    def add_file(self, path=None):
        if path is None:
            new_path = QFileDialog.getOpenFileName(
                self,
                "Open File",
                "${HOME}",
                "NIfTI Compressed Files (*.nii.gz);; NIfTI Files (*.nii)",
            )[0]
        else: new_path = path

        if not hasattr(self, "nifti_found"):
            self.nifti_found = []
        if new_path not in self.nifti_found:
            self.nifti_found.append(new_path)
            if not hasattr(self, "thumbnail_title"): 
                self.init_all()
                self.load_sequences(self.nifti_found)
            else: self.load_sequence(new_path)
            self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        self.add_recent_path(new_path)
        
    def load_sequences(self, paths):
        self.thumbnail_widget = QWidget()
        self.thumbnail_layout = QVBoxLayout()
        self.thumbnail_layout.addStretch()

        self.thumbnail_title = QLabel()
        self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(paths)} images loaded:</h2>")
        self.thumbnail_layout.addWidget(self.thumbnail_title)

        if len(paths) > 1:
            data_loader = DataLoader(self.thumbnail_layout, self.replace_data, self.create_overlay)
            for path in paths:
                print(path)
                data_loader.load_sequence(path)
        else:
            self.load_sequence(paths[0])

        self.thumbnail_widget.setLayout(self.thumbnail_layout)
        self.scroll_area.setWidget(self.thumbnail_widget)

        print("INFO - UI fully created")

    def load_sequence(self, path):
        print(f"INFO - Found:\t{path}")
        frame = ThumbnailFrame(path, self.replace_data, self.create_overlay)
        label_image = QLabel()

        thumbnail, shape, pixdim = create_thumbnail(path)
        thumbnail = (thumbnail - thumbnail.min()) / (thumbnail.max() - thumbnail.min()) * 255
        thumbnail = thumbnail.astype('uint8')
        image = QImage(thumbnail.data, thumbnail.shape[1], thumbnail.shape[0], thumbnail.shape[1], QImage.Format.Format_Grayscale8)

        pixmap = QPixmap.fromImage(image)
        pixel_width, pixel_height, _ = pixdim
        scaled_width = int(pixmap.width() * pixel_width)
        scaled_height = int(pixmap.height() * pixel_height)
        pixmap = pixmap.scaled(scaled_width, scaled_height, Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation)
        pixmap = pixmap.scaledToWidth(THUMBNAIL_WIDTH, Qt.TransformationMode.SmoothTransformation)
        label_image.setPixmap(pixmap)

        label_info = QLabel()
        path = os.path.basename(path).split(".nii")[0]
        if len(path) > 32: path = path[:32] + "..."
        pixdim = [round(pixdim[0], 2), round(pixdim[1], 2), round(pixdim[2], 2)]
        label_info.setText(f"{path}<br>Shape: {shape}<br>Pixdim: {pixdim}<hr style='background-color: {LIGHT_BORDER_COLOR}'>")

        label_info.setStyleSheet("border: 0px solid #000000;")
        label_image.setStyleSheet("border: 0px solid #000000;")

        temp_layout = QVBoxLayout(frame)
        temp_layout.addWidget(label_image)
        temp_layout.addWidget(label_info)

        self.thumbnail_layout.addWidget(frame)

    def replace_data(self, path):
        self.view_widget.init(path)
        self.view_widget.update_text()
        self.remove_overlay()

    def create_overlay(self, path):
        if self.overlay: self.remove_overlay()
        self.overlay = True
        self.overlay_path = path

        self.overlay_label = QLabel("<h2><b>Overlay:</b></h2>")

        self.overlay_toggle = QWidget()
        self.overlay_toggle_layout = QHBoxLayout()
        self.overlay_toggle_label = QLabel("Toggle overlay visibility")
        self.overlay_toggle_button = AnimatedToggle(self.toggle_overlay_visibility)
        self.overlay_toggle_layout.addWidget(self.overlay_toggle_label)
        self.overlay_toggle_layout.addWidget(self.overlay_toggle_button)
        self.overlay_toggle.setLayout(self.overlay_toggle_layout)
        
        self.overlay_slider = QWidget()
        self.overlay_slider_layout = QHBoxLayout()
        self.overlay_slider_label = QLabel("Overlay opacity")
        self.overlay_slider_button = QSlider(Qt.Orientation.Horizontal, self)
        self.overlay_slider_button.setGeometry(50,50, 200, 50)
        self.overlay_slider_button.setMinimum(0)
        self.overlay_slider_button.setMaximum(100)
        self.overlay_slider_button.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.overlay_slider_button.setValue(50)
        self.overlay_slider_button.setTickInterval(1)
        self.overlay_slider_button.valueChanged.connect(self.change_overlay_opacity) 
        self.overlay_slider_value = QLabel("50%")
        self.overlay_slider_layout.addWidget(self.overlay_slider_label)
        self.overlay_slider_layout.addWidget(self.overlay_slider_button)
        self.overlay_slider_layout.addWidget(self.overlay_slider_value)
        self.overlay_slider.setLayout(self.overlay_slider_layout)

        self.overlay_lut = QWidget()
        self.overlay_lut_layout = QVBoxLayout()
        self.overlay_lut_label = QLabel("Look-up-table")
        self.overlay_lut_button = ComboBox()
        self.overlay_lut_button.addItems(['Grayscale', 'Red', 'Green', 'Blue', 'Rainbow', 'Spectral', 'Flow', 'Viridis', 'Qualitative', 'Random'])
        self.overlay_lut_button.currentTextChanged.connect(self.change_overlay_lut)
        self.overlay_lut_layout.addWidget(self.overlay_lut_label)
        self.overlay_lut_layout.addWidget(self.overlay_lut_button)
        self.overlay_lut.setLayout(self.overlay_lut_layout)

        self.overlay_remove = QPushButton("Remove overlay")
        self.overlay_remove.setFixedSize(150, 60)
        self.overlay_remove.setDown(True)
        self.overlay_remove.clicked.connect(self.remove_overlay)

       # Insert the overlay label and button at the top of the layout
        self.info_layout.insertWidget(2, self.overlay_label)
        self.info_layout.insertWidget(3, self.overlay_toggle)
        self.info_layout.insertWidget(4, self.overlay_slider)
        self.info_layout.insertWidget(5, self.overlay_lut)
        self.info_layout.insertWidget(6, self.overlay_remove, alignment=Qt.AlignCenter)
        self.info_layout.addStretch()

        self.view_widget.add_overlay(self.overlay_path)

    def change_lut(self, lut):
        self.view_widget.change_lut(lut)

    def change_overlay_lut(self, lut):
        self.view_widget.change_overlay_lut(lut)

    def toggle_overlay_visibility(self):
        self.view_widget.layout.itemAt(0).widget().overlay_toggle = not self.view_widget.layout.itemAt(0).widget().overlay_toggle
        self.view_widget.layout.itemAt(0).widget().update_image()
        if self.view_widget.state == "3D":
            self.view_widget.layout.itemAt(1).widget().overlay_toggle = not self.view_widget.layout.itemAt(1).widget().overlay_toggle
            self.view_widget.layout.itemAt(2).widget().overlay_toggle = not self.view_widget.layout.itemAt(2).widget().overlay_toggle
            self.view_widget.layout.itemAt(1).widget().update_image()
            self.view_widget.layout.itemAt(2).widget().update_image()

    def change_overlay_opacity(self):
        self.overlay_slider_value.setText(f"{self.sender().value()}%")
        for i in reversed(range(self.view_widget.layout.count())):
            label = self.view_widget.layout.itemAt(i).widget()
            label.overlay_opacity = self.sender().value()
            label.update_image()

    def remove_overlay(self):
        if self.overlay:
            for i in reversed(range(self.view_widget.layout.count())):
                label = self.view_widget.layout.itemAt(i).widget()
                label.overlay = False
                label.overlay_toggle = False
                label.overlay_data = None
                label.update_image()

            self.info_layout.removeWidget(self.overlay_label)
            self.info_layout.removeWidget(self.overlay_toggle)
            self.info_layout.removeWidget(self.overlay_slider)
            self.info_layout.removeWidget(self.overlay_lut)
            self.info_layout.removeWidget(self.overlay_remove)
            self.overlay_label.setParent(None)
            self.overlay_toggle.setParent(None)
            self.overlay_slider.setParent(None)
            self.overlay_lut.setParent(None)
            self.overlay_remove.setParent(None)

        self.overlay = False
        self.overlay_path = None

    def screenshot(self):
        screen = QApplication.primaryScreen()
        dir = QFileDialog.getExistingDirectory(self, "Select a directory", "C:\\", QFileDialog.ShowDirsOnly)

        if self.view_widget.state != "3D":
            if self.view_widget.state == "Axial":
                screenshot = screen.grabWindow(self.view_widget.axial_label.winId())
                name = f"{os.path.basename(self.view_widget.path).replace('.nii', '').replace('.gz', '')}_axial.png"
            elif self.view_widget.state == "Coronal":
                screenshot = screen.grabWindow(self.view_widget.coronal_label.winId())
                name = f"{os.path.basename(self.view_widget.path).replace('.nii', '').replace('.gz', '')}_coronal.png"
            elif self.view_widget.state == "Sagittal":
                screenshot = screen.grabWindow(self.view_widget.sagittal_label.winId())
                name = f"{os.path.basename(self.view_widget.path).replace('.nii', '').replace('.gz', '')}_sagittal.png"

            screenshot.save(os.path.join(dir, name), "png")
            print(f"INFO - Successfully screenshot at: {os.path.join(dir, name)}")
        else:
            screenshot = screen.grabWindow(self.view_widget.axial_label.winId())
            name = f"{os.path.basename(self.view_widget.path).replace('.nii', '').replace('.gz', '')}_axial.png"
            screenshot.save(os.path.join(dir, name), "png")

            screenshot = screen.grabWindow(self.view_widget.coronal_label.winId())
            name = f"{os.path.basename(self.view_widget.path).replace('.nii', '').replace('.gz', '')}_coronal.png"
            screenshot.save(os.path.join(dir, name), "png")

            screenshot = screen.grabWindow(self.view_widget.sagittal_label.winId())
            name = f"{os.path.basename(self.view_widget.path).replace('.nii', '').replace('.gz', '')}_sagittal.png"
            screenshot.save(os.path.join(dir, name), "png")
        
    def gifshot(self):
        gif_array = []
        current_slice = self.view_widget.a_slice
        self.view_widget.axial_label.data = self.view_widget.adata[..., self.view_widget.a_slice]
        for slice in range(self.view_widget.shape[2]):
            print(slice)
            self.view_widget.update_slice(slice, "Axial")
            self.view_widget.axial_label.update_image()

            pixmap = self.view_widget.axial_label.pixmap

            # Convert the pixmap to a QImage
            image = pixmap.toImage()

            # Convert the QImage to a NumPy array
            width = image.width()
            height = image.height()
            buffer = image.constBits()
            dtype = np.uint8 if image.format() == QImage.Format.Format_RGB32 else np.uint32
            array = np.array(buffer).reshape(height, width, 4).astype(dtype)

            array = array[:, :, [2, 1, 0, 3]]
            gif_array.append(array)  # Extract RGB channels

        print("INFO - Making the gif file...")
        imageio.mimsave(f'{self.dirname}/animation.gif', gif_array, duration=0.375/(len(gif_array)/10))
        self.view_widget.update_slice(current_slice, "Axial")
        self.view_widget.axial_label.update_image()
 
    def whole_histogram(self):
        # Flatten the 3D array to a 1D array
        image_flat = self.rotated_data.flatten()
        image_flat = np.where(image_flat == 0, np.NaN, image_flat)

        # Plot the histogram
        plt.hist(image_flat, bins=256, range=(1, np.amax(self.rotated_data)), alpha=0.75)
        plt.xlabel('Intensity')
        plt.xlim(left=1)
        plt.ylabel('Frequency')
        plt.title('Histogram of the whole image')
        plt.show()

    def toggle_view(self, state):
        self.view_widget.state = state
        reset_layout(self.view_widget.layout)
        self.view_widget.update_layout()
        self.view_widget.update_text()

    def toggle_mosaic(self):
        self.reset_layout(self.view_widget.layout)
        self.create_image_label(self.filename, "mosaic")

    def add_current_histogram(self):
        self.toggle_current_histo = not self.toggle_current_histo
        if not self.toggle_current_histo:
            self.info_layout.removeWidget(self.graphWidget_2)
            self.graphWidget_2.setParent(None)
        else:
            if self.info_layout.count() <= 1:
                self.update_histogram()
                self.info_layout.addWidget(self.graphWidget_2, 1,0) 

    def update_aif(self, aif):
        self.plot_aif.clear()
        self.plot_aif.plot(aif)

    def create_histogram(self):
        image_flat = self.data.flatten()
        counts, bins = np.histogram(image_flat,bins=256)
        counts = counts[1:]

        self.graphWidget_1 = pg.PlotWidget()
        self.graphWidget_1.plot(np.arange(0, len(counts), 1), counts, pen=pen)
        self.graphWidget_2 = pg.PlotWidget()

        self.info_layout.addWidget(self.graphWidget_1, 1,0)

    def update_histogram(self):
        self.graphWidget_2.clear()
        image_flat = self.rotated_data[:, :, self.slice_index].flatten()
        counts, bins = np.histogram(image_flat,bins=256)
        counts = counts[1:]

        self.graphWidget_2.plot(np.arange(0, len(counts), 1), counts, pen=pen) 

    def update_text(self):
        title = "<h2><b>Information:</b></h2>"
        info = f"{os.path.basename(self.view_widget.path)}<br>"
        info += f"Image dimensions: {self.view_widget.shape}<br>Pixel dimension: {self.view_widget.pixdim}<br>"
        if self.view_widget.state == "3D" :
            info += f"X: {self.view_widget.s_slice}<br>"
            info += f"Y: {self.view_widget.c_slice}<br>"
            info += f"Z: {self.view_widget.a_slice}<br>"
        else:
            # if hasattr(self.view_widget, "layout") and self.view_widget.layout.count() > 0:
                image_label = self.view_widget.get_state_label()
                if len(self.view_widget.shape) == 4:
                    phase_index = image_label.slice_index // self.view_widget.shape[2]
                    slice_index = image_label.slice_index % self.view_widget.shape[2]
                    if self.view_widget.state == "Mosaic":
                        info += f"Phase: {image_label.slice_index+1}<br><br>"
                    else:
                        info += f"Phase: {phase_index+1}<br>Slice: {slice_index+1}<br><br>"
                else: 
                    info += f"Slice: {image_label.slice_index+1}<br>"
                    info += f"Window width: {int(image_label.window_width)}<br>Window center: {int(image_label.window_center)}"
        self.text_label.setText(f"{title}{info}")
        self.roi_z.setText(f"{self.view_widget.a_slice}")

    def update_roi(self, x, y, z, mean, std, min, max, mm3):
        self.roi_x.setText(str(int(x)))
        self.roi_y.setText(str(int(y)))
        self.roi_z.setText(str(int(z)))
        self.roi_table.setItem(0, 0, QTableWidgetItem(str(round(mean,3))))
        self.roi_table.setItem(0, 1, QTableWidgetItem(str(round(std,3))))
        self.roi_table.setItem(0, 2, QTableWidgetItem(str(round(min,3))))
        self.roi_table.setItem(0, 3, QTableWidgetItem(str(round(max,3))))
        self.roi_table.setItem(0, 4, QTableWidgetItem(str(round(mm3,3))))

    def reset_contrast(self):
        for i in reversed(range(self.view_widget.layout.count())):
            widget = self.view_widget.layout.itemAt(i).widget()
            widget.window_center = (widget.contrast_min + widget.contrast_max) / 2  # Reinitial contrast center
            widget.window_width = widget.contrast_max - widget.contrast_min  # Reinitial contrast width
            widget.update_image()

    def transforms(self, transform):
        if "DICOM" in transform: 
            path = None
            widget_path = None
        else: 
            path = self.nifti_found
            widget_path = self.view_widget.path
        new_window = TransformsGUI(path, widget_path, transform, self.run_transform, self.init_config)
        new_window.show()
        self.dynamic_windows.append(new_window)

    def run_transform(self, *args):
        if args[0] == "Sort DICOM":
            sort_dicom(args[1], args[2], anonymize=False)
        elif args[0] == "Convert DICOM":
            convert_dicom_to_nifti(args[1], args[2])
        elif args[0] == "Auto crop":
            new_path, idx = crop(args[1], force=args[2])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Apply crop":
            idx = [int(val) for val in args[2].split(",")]
            new_path = apply_crop(args[1], crop=idx, force=args[3])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Resample voxels":
            new_path = resample(args[1], float(args[2].replace(",", ".")))[0]
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Remove islands":
            new_path = remove_small_object(args[1], int(args[2]), force=args[3])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Susan":
            new_path = susan(args[1], offset=int(args[2]), force=args[3])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Registration":
            if args[3] is True:
                new_path, _ = registration(args[1], args[2], matrix=args[3], force=args[4])
            else:
                new_path = registration(args[1], args[2], matrix=args[3], force=args[4])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Apply transformation":
            new_path = apply_transformation(args[1], args[2], args[3], force=args[3])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Apply mask":
            new_path = apply_mask(args[1], args[2], force=args[3])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Skull stripping":
            new_path2 = None
            if args[3] is True:
                new_path, new_path2 = skull_stripping(args[1], model_path=args[2], mask=args[3], force=args[4])
            else:
                new_path = skull_stripping(args[1], model_path=args[2], mask=args[3], force=args[4])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
            if new_path2 is not None and new_path2 not in self.nifti_found:
                self.load_sequence(new_path2)
                self.nifti_found.append(new_path2)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Bias field correction":
            new_path2 = None
            if args[3] is True:
                new_path, new_path2 = n4_bias_field_correction(args[1], mask=args[2], force=args[4])
            else:
                new_path = n4_bias_field_correction(args[1], force=args[2])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
            if new_path2 is not None and new_path2 not in self.nifti_found:
                self.load_sequence(new_path2)
                self.nifti_found.append(new_path2)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Glioma":
            new_path = prediction_glioma(args[1], model_path=args[2], landmark_path=args[3], force=args[4])
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
        elif args[0] == "Multiple Sclerosis":
            new_path, new_path2 = prediction_multiple_sclerosis(args[1], model_path=args[2], landmark_path=args[3], force=args[4])
            print(new_path)
            if new_path not in self.nifti_found:
                self.load_sequence(new_path)
                self.nifti_found.append(new_path)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")
            if new_path2 not in self.nifti_found:
                self.load_sequence(new_path2)
                self.nifti_found.append(new_path2)
                self.thumbnail_title.setText(f"<h2 style='margin-left: 10px'>{len(self.nifti_found)} images loaded:</h2>")

    def closeEvent(self, event):
        for window in self.dynamic_windows:
            window.close()
        event.accept()

    def center(self):
        qr = self.frameGeometry()
        cp = self.screen().availableGeometry().center()

        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def dragEnterEvent(self, event: QDragEnterEvent):
        if event.mimeData().hasUrls():
            event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent):
        if event.mimeData().hasUrls():
            event.setDropAction(Qt.DropAction.CopyAction)
            event.accept()
            for url in event.mimeData().urls():
                file_path = url.toLocalFile()
                self.handleFileDrop(file_path)
    
    def handleFileDrop(self, path):
        # Call your function here with the path
        print(f"Dropped: {path}")
        if os.path.isdir(path):
            self.open_volume(path=path)
        else:
            self.add_file(path=path)

class WorkerSignals(QObject):
    finished = Signal()
    result = Signal(object)

class LoadSequenceWorker(QRunnable):
    def __init__(self, path, callback):
        super().__init__()
        self.path = path
        self.callback = callback
        self.signals = WorkerSignals()

    @Slot()
    def run(self):
        try:
            thumbnail, shape, pixdim = create_thumbnail(self.path)
            thumbnail = (thumbnail - thumbnail.min()) / (thumbnail.max() - thumbnail.min()) * 255
            thumbnail = thumbnail.astype('uint8')
            self.signals.result.emit((self.path, thumbnail, shape, pixdim))
        except Exception as e:
            print(f"Error in worker for {self.path}: {e}")
        finally:
            self.signals.finished.emit()

class DataLoader(QObject):
    def __init__(self, thumbnail_layout, replace_data, create_overlay):
        super().__init__()
        self.thumbnail_layout = thumbnail_layout
        self.threadpool = QThreadPool()
        self.active_workers = 0
        self.finished_workers = 0
        self.replace_data = replace_data
        self.create_overlay = create_overlay

    def load_sequence(self, path):
        worker = LoadSequenceWorker(path, self.process_thumbnail)
        worker.signals.result.connect(self.process_thumbnail)
        worker.signals.finished.connect(self.worker_finished)
        self.active_workers += 1
        self.threadpool.start(worker)
        
    @Slot(object)
    def process_thumbnail(self, result):
        try:
            path, thumbnail, shape, pixdim = result
            frame = ThumbnailFrame(path, self.replace_data, self.create_overlay)
            label_image = QLabel()

            image = QImage(thumbnail.data, thumbnail.shape[1], thumbnail.shape[0], thumbnail.shape[1], QImage.Format.Format_Grayscale8)
            pixmap = QPixmap.fromImage(image)
            pixel_width, pixel_height, _ = pixdim
            scaled_width = int(pixmap.width() * pixel_width)
            scaled_height = int(pixmap.height() * pixel_height)
            pixmap = pixmap.scaled(scaled_width, scaled_height, Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation)
            pixmap = pixmap.scaledToWidth(THUMBNAIL_WIDTH, Qt.TransformationMode.SmoothTransformation)
            label_image.setPixmap(pixmap)

            label_info = QLabel()
            path = os.path.basename(path).split(".nii")[0]
            if len(path) > 32: path = path[:32] + "..."
            pixdim = [round(pixdim[0], 2), round(pixdim[1], 2), round(pixdim[2], 2)]
            label_info.setText(f"{path}<br>Shape: {shape}<br>Pixdim: {pixdim}<hr style='background-color: {LIGHT_BORDER_COLOR}'>")

            label_info.setStyleSheet("border: 0px solid #000000;")
            label_image.setStyleSheet("border: 0px solid #000000;")

            temp_layout = QVBoxLayout(frame)
            temp_layout.addWidget(label_image)
            temp_layout.addWidget(label_info)

            self.thumbnail_layout.addWidget(frame)
        except Exception as e:
            print(f"Error processing thumbnail: {e}")

    @Slot()
    def worker_finished(self):
        self.finished_workers += 1
        if self.finished_workers + 1 == self.active_workers:
            self.threadpool.waitForDone()

class TransformsGUI(QWidget):
    def __init__(self, path, selected_path, transform, run_transform, conf):
        super().__init__()
        layout = QVBoxLayout()
        self.setLayout(layout)

        self.setWindowTitle("PyDactim Transformation")
        self.show_centered()

        settings = conf()

        self.path = path
        self.run_transform = run_transform

        if transform == "Sort DICOM":
            title = QLabel("Path to the DICOM directory")
            input_dir = QPushButton("Browse")
            input_dir.clicked.connect(self.get_input_dir)
            self.input_dir = QLineEdit(self)
            layout.addWidget(title)
            layout.addWidget(input_dir)
            layout.addWidget(self.input_dir)

            title = QLabel("Path to the DICOM sorted directory")
            output_dir = QPushButton("Browse")
            output_dir.clicked.connect(self.get_output_dir)
            self.output_dir = QLineEdit(self)
            layout.addWidget(title)
            layout.addWidget(output_dir)
            layout.addWidget(self.output_dir)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, self.input_dir.text(), self.output_dir.text()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Convert DICOM":
            title = QLabel("Path to the DICOM directory")
            input_dir = QPushButton("Browse")
            input_dir.clicked.connect(self.get_input_dir)
            self.input_dir = QLineEdit(self)
            layout.addWidget(title)
            layout.addWidget(input_dir)
            layout.addWidget(self.input_dir)

            title = QLabel("Path to the converted NIfTI file(s)")
            output_dir = QPushButton("Browse")
            output_dir.clicked.connect(self.get_output_dir)
            self.output_dir = QLineEdit(self)
            layout.addWidget(title)
            layout.addWidget(output_dir)
            layout.addWidget(self.output_dir)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, self.input_dir.text(), self.output_dir.text()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Auto crop":
            title = QLabel("Path for the image to auto crop")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)
        
        elif transform == "Apply crop":
            title = QLabel("Path for for the image to apply crop")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Indices")
            idx = QLineEdit(self)
            idx.setPlaceholderText("x1, x2, y1, y2, z1, z2")
            layout.addWidget(title)
            layout.addWidget(idx)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), idx.text(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Resample voxels":
            title = QLabel("Path for image the to resample")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Voxel dimension (mm)")
            voxel = QLineEdit(self)
            voxel.setText("1")
            layout.addWidget(title)
            layout.addWidget(voxel)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), voxel.text()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Remove islands":
            title = QLabel("Path for the image in which removing islands")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Minimum size of objects to be removed")
            voxel = QLineEdit(self)
            layout.addWidget(title)
            layout.addWidget(voxel)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), voxel.text(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Susan":
            title = QLabel("Path for the image to susan")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Offset")
            offset = QLineEdit(self)
            offset.setText("3")
            layout.addWidget(title)
            layout.addWidget(offset)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), offset.text(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Skull stripping":
            title = QLabel("Path for the image to skull strip")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Model path")
            dialog = QPushButton("Browse")
            dialog.clicked.connect(self.get_extra_dir)
            self.extra = QLineEdit(self)
            self.extra.setText(settings.value("ai_model_paths/skull_stripping", ""))
            
            layout.addWidget(title)
            layout.addWidget(self.extra)
            layout.addWidget(dialog)

            mask = QCheckBox("Mask")
            layout.addWidget(mask)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), self.extra.text(), mask.isChecked(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Registration":
            title = QLabel("Path for fixed image")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Path for moving image")
            combo_box_2 = QComboBox(self)
            combo_box_2.addItems(path)
            combo_box_2.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box_2)

            mat = QCheckBox("Matrix")
            layout.addWidget(mat)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), combo_box_2.currentText(), mat.isChecked(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Apply transformation":
            title = QLabel("Path for fixed image")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Path for moving image")
            combo_box_2 = QComboBox(self)
            combo_box_2.addItems(path)
            combo_box_2.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box_2)

            title = QLabel("Matrix path")
            dialog = QPushButton("Browse")
            dialog.clicked.connect(self.get_extra_file)
            self.extra = QLineEdit(self)
            
            layout.addWidget(title)
            layout.addWidget(self.extra)
            layout.addWidget(dialog)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), combo_box_2.currentText(), self.extra.text(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Apply mask":
            title = QLabel("Path for the image to be masked")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Path for the mask")
            combo_box_2 = QComboBox(self)
            combo_box_2.addItems(path)
            combo_box_2.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box_2)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), combo_box_2.currentText(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Bias field correction":
            title = QLabel("Path for the image to correct")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            mask = QCheckBox("Mask")
            layout.addWidget(mask)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), mask.isChecked(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Glioma":
            title = QLabel("Path for the input T2-FLAIR")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Model path (including landmarks)")
            dialog = QPushButton("Browse")
        
            dialog.clicked.connect(self.get_ai_dir)
            self.model = QLineEdit(self)
            self.extra = QLineEdit(self)

            path = settings.value("ai_model_paths/glioma", "")
            if path is not None: self.set_ai_dir(path)
            else: "C:\\"

            layout.addWidget(title)
            layout.addWidget(dialog)
            layout.addWidget(self.model)
            layout.addWidget(self.extra)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), self.model.text(), self.extra.text(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

        elif transform == "Multiple Sclerosis":
            title = QLabel("Path for the input T2-FLAIR")
            combo_box = QComboBox(self)
            combo_box.addItems(path)
            combo_box.setCurrentText(selected_path)
            layout.addWidget(title)
            layout.addWidget(combo_box)

            title = QLabel("Model path (including landmarks)")
            dialog = QPushButton("Browse")
        
            dialog.clicked.connect(self.get_ai_dir)
            self.model = QLineEdit(self)
            self.extra = QLineEdit(self)

            path = settings.value("ai_model_paths/ms", "")
            if path is not None and len(path) > 0: self.set_ai_dir(path)
            else: "C:\\"

            layout.addWidget(title)
            layout.addWidget(dialog)
            layout.addWidget(self.model)
            layout.addWidget(self.extra)

            force = QCheckBox("Force")
            layout.addWidget(force)

            button = QPushButton("Run Function", self)
            button.clicked.connect(lambda: self.launch_transform(transform, combo_box.currentText(), self.model.text(), self.extra.text(), force.isChecked()))
            layout.addWidget(button, alignment=Qt.AlignmentFlag.AlignHCenter)

    def set_ai_dir(self, dir):
        for file in os.listdir(dir):
            if file.endswith(".pth"): self.model.setText(os.path.join(dir, file))
            elif file.endswith(".npy"): self.extra.setText(os.path.join(dir, file))

    def get_ai_dir(self):
        dialog = QFileDialog()
        dir = dialog.getExistingDirectory(None, "Select a directory", "C:\\", QFileDialog.ShowDirsOnly)
        self.set_ai_dir(dir)

    def get_extra_dir(self):
        dialog = QFileDialog()
        dir = dialog.getExistingDirectory(None, "Select a directory", "C:\\", QFileDialog.ShowDirsOnly)
        self.extra.setText(dir)

    def get_extra_file(self):
        file = QFileDialog.getOpenFileName(
            self,
            "Open File",
            os.path.dirname(self.path[0]),
            "Matrix files (*.tfm);;",
        )[0]
        self.extra.setText(file)

    def get_input_dir(self):
        dialog = QFileDialog()
        dir = dialog.getExistingDirectory(None, "Select a directory", "C:\\", QFileDialog.ShowDirsOnly)
        self.input_dir.setText(dir)

    def get_output_dir(self):
        dialog = QFileDialog()
        dir = dialog.getExistingDirectory(None, "Select a directory", "C:\\", QFileDialog.ShowDirsOnly)
        self.output_dir.setText(dir)

    def launch_transform(self, *args):
        self.run_transform(*args)
        self.close()
    
    def show_centered(self):
        app = QGuiApplication.instance()
        screen_geometry = app.primaryScreen().availableGeometry()
        center = screen_geometry.center()
        self.move(center.x() - self.width() / 2, center.y() - self.height() / 2)
        self.show()

class CustomOutputStream(StringIO):
    def __init__(self, label, scroll):
        super().__init__()
        self.label = label
        self.scroll = scroll

    def write(self, txt):
        super().write(txt)
        self.label.setText(self.getvalue())
        self.scroll.verticalScrollBar().setValue(
            self.scroll.verticalScrollBar().maximum()
        )

def ViewerApp():
    app = QApplication(sys.argv + ['-platform', 'windows:darkmode=2'])
    app.setStyle('Fusion')
    app.setPalette(get_darkModePalette(app))

    # default_nifti_file = "path/to/your/default/nifti_file.nii.gz"  # Change this to the default NIfTI file path

    # if not os.path.isfile(default_nifti_file):
    #     default_nifti_file, _ = QFileDialog.getOpenFileName(None, "Open NIfTI file", "", "NIfTI Files (*.nii.gz)")

    # if not os.path.isfile(default_nifti_file):
    #     print("Error: No valid NIfTI file selected.")
    #     sys.exit(1)

    viewer = NiftiViewer()
    viewer.show()
    viewer.center()
    sys.exit(app.exec())

if __name__ == "__main__":
    ViewerApp()




from PySide6.QtGui import QColor, QPalette
from PySide6.QtCore import Qt
import matplotlib.pyplot as plt
import nibabel as nib
from numba import njit
import numpy as np

def is_dsc_param(path):
        if "CBV" in path.upper() or "CBF" in path.upper() or "MTT" in path.upper() or "TTP" in path.upper():
            return True
        return False

def to_sagittal(data, pixdim, path):
    sdata = np.transpose(data, (1,2,0))
    spixdim = [pixdim[1], pixdim[2], pixdim[0]]
    sdata = np.rot90(sdata) if "anat" in path else np.rot90(sdata[:,:,::-1])
    sdata = np.fliplr(sdata)
    s_slice = sdata.shape[2] // 2
    return sdata, s_slice, spixdim, sdata.shape[2]

def to_coronal(data, pixdim, path):
    cdata = np.transpose(data, (0,2,1))
    cpixdim = [pixdim[0], pixdim[2], pixdim[1]]
    cdata = np.rot90(cdata[:,:,::-1])
    if "anat" in path:
        cdata = np.fliplr(cdata)
    elif "dwi" in path and "flirt" in path:
        cdata = np.fliplr(cdata)
    elif "perf" in path and "flirt" in path:
        cdata = np.fliplr(cdata)
    c_slice = cdata.shape[2] // 2
    return cdata, c_slice, cpixdim, cdata.shape[2]

def to_axial(data, pixdim, path):
    adata = np.rot90(data)
    apixdim = list(pixdim)
    if "anat" in path and not "flirt" in path:
        adata = np.fliplr(adata) 
    elif "dwi" in path and "flirt" in path:
        adata = np.fliplr(adata)
    elif "perf" in path and "flirt" in path:
        adata = np.fliplr(adata)
    a_slice = adata.shape[2] // 2
    return adata, a_slice, apixdim, adata.shape[2]

def reset_layout(layout):
    # Clear the layout
    for i in reversed(range(layout.count())):
        widget = layout.itemAt(i).widget()
        if widget is not None:
            layout.removeWidget(widget)
            widget.setParent(None)

def create_lut(lut):
    if lut == "Grayscale":
        color_lut = np.zeros((256, 3), dtype=np.uint8)
        color_lut[:, 0] = np.arange(256)
        color_lut[:, 1] = np.arange(256)
        color_lut[:, 2] = np.arange(256)
    elif lut == "Red":
        color_lut = np.zeros((256, 3), dtype=np.uint8)
        color_lut[:, 0] = np.arange(256)
        color_lut[:, 1] = 0
        color_lut[:, 2] = 0
    elif lut == "Green":
        color_lut = np.zeros((256, 3), dtype=np.uint8)
        color_lut[:, 0] = 0
        color_lut[:, 1] = np.arange(256)
        color_lut[:, 2] = 0
    elif lut == "Blue":
        color_lut = np.zeros((256, 3), dtype=np.uint8)
        color_lut[:, 0] = 0
        color_lut[:, 1] = 0
        color_lut[:, 2] = np.arange(256)
    elif lut == "Rainbow":
        colormap = plt.get_cmap("jet", 256)
        color_lut = (colormap(np.arange(256))[:, :3] * 255).astype(np.uint8)
    elif lut == "Spectral":
        colormap = plt.get_cmap("nipy_spectral", 256)
        color_lut = (colormap(np.arange(256))[:, :3] * 255).astype(np.uint8)
    elif lut == "Flow":
        colormap = plt.get_cmap("coolwarm", 256)
        color_lut = (colormap(np.arange(256))[:, :3] * 255).astype(np.uint8)
    elif lut == "Viridis":
        colormap = plt.get_cmap("viridis", 256)
        color_lut = (colormap(np.arange(256))[:, :3] * 255).astype(np.uint8)
    elif lut == "Qualitative":
        colormap = plt.get_cmap("Paired", 256)
        color_lut = (colormap(np.arange(256))[:, :3] * 255).astype(np.uint8)
    elif lut == "Random":
        colormap = plt.get_cmap("prism", 256)
        color_lut = (colormap(np.arange(256))[:, :3] * 255).astype(np.uint8)
    return color_lut

def create_thumbnail(path):
    img = nib.load(path)
    data = np.array(img.dataobj, dtype=np.float32)

    def thumbnail(data):
        if len(data.shape) == 4:
            data = data.transpose(0, 1, 3, 2).reshape(
                data.shape[0], 
                data.shape[1], 
                data.shape[2]*data.shape[3]
            )
            
        data = np.rot90(data)
        if "anat" in path:
            data = np.fliplr(data)
        elif "dwi" in path and "flirt" in path:
            data = np.fliplr(data)
        elif "perf" in path and "flirt" in path:
            data = np.fliplr(data)

        slice_index = data.shape[2] // 2
        data = data[..., slice_index]
        return data
        
    data = thumbnail(data)
    return data, list(img.shape), list(img.header["pixdim"][1:4])

def to_3D(array, pixdim):
    array = np.rot90(array)
    array = np.transpose(array, (0, 1, 3, 2))
    array = np.reshape(array,
        (array.shape[0], 
        array.shape[1], 
        array.shape[2]*array.shape[3])
    )
    pixdim = [pixdim[1], pixdim[0], pixdim[2]]
    return array, pixdim, array.shape[2]

@njit
def create_3D_grid(array):
    num_slices = array.shape[2]
    num_rows = int(np.ceil(np.sqrt(num_slices)))  # Number of rows in the grid
    num_cols = int(np.ceil(num_slices / num_rows))  # Number of columns in the grid

    # Create a blank canvas to store the slices
    canvas_height = array.shape[0] * num_rows
    canvas_width = array.shape[1] * num_cols
    canvas = np.zeros((canvas_height, canvas_width), dtype=array.dtype)

    # Fill the canvas with the slices
    for i in range(num_slices):
        row = i // num_cols
        col = i % num_cols
        canvas[row * array.shape[0]:(row + 1) * array.shape[0], col * array.shape[1]:(col + 1) * array.shape[1]] = array[:, :, i]
    return canvas

@njit
def create_4D_grid(array):
    array = np.rot90(array)
    num_slices = array.shape[2]
    num_rows = int(np.ceil(np.sqrt(num_slices)))  # Number of rows in the grid
    num_cols = int(np.ceil(num_slices / num_rows))  # Number of columns in the grid

    # Create a blank canvas to store the slices
    canvas_height = array.shape[0] * num_rows
    canvas_width = array.shape[1] * num_cols
    canvas = np.zeros((canvas_height, canvas_width, array.shape[3]), dtype=array.dtype)

    # Fill the canvas with the slices
    for i in range(array.shape[3]):
        temp = create_3D_grid(array[:,:,:,i])
        canvas[:,:,i] = temp
    return canvas

def get_darkModePalette(app=None):
    darkPalette = app.palette()
    darkPalette.setColor(QPalette.Window, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.WindowText, Qt.white)
    darkPalette.setColor(QPalette.Disabled, QPalette.WindowText, QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Base, QColor(42, 42, 42))
    darkPalette.setColor(QPalette.AlternateBase, QColor(66, 66, 66))
    darkPalette.setColor(QPalette.ToolTipBase, Qt.white)
    darkPalette.setColor(QPalette.ToolTipText, Qt.white)
    darkPalette.setColor(QPalette.Text, Qt.white)
    darkPalette.setColor(QPalette.Disabled, QPalette.Text, QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Dark, QColor(35, 35, 35))
    darkPalette.setColor(QPalette.Shadow, QColor(20, 20, 20))
    darkPalette.setColor(QPalette.Button, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.ButtonText, Qt.white)
    darkPalette.setColor(QPalette.Disabled, QPalette.ButtonText, QColor(127, 127, 127))
    darkPalette.setColor(QPalette.BrightText, Qt.red)
    darkPalette.setColor(QPalette.Link, QColor(42, 130, 218))
    darkPalette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    darkPalette.setColor(QPalette.Disabled, QPalette.Highlight, QColor(80, 80, 80))
    darkPalette.setColor(QPalette.HighlightedText, Qt.white)
    darkPalette.setColor(QPalette.Disabled, QPalette.HighlightedText, QColor(127, 127, 127),)
    return darkPalette